# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Disable logging for a cleaner testing output
import logging

from helpers import AppriseURLTester

from apprise.plugins.bark import NotifyBark

logging.disable(logging.CRITICAL)

# Our Testing URLs
apprise_url_tests = (
    (
        "bark://",
        {
            # No no host
            "instance": None,
        },
    ),
    (
        "bark://:@/",
        {
            # just invalid all around
            "instance": None,
        },
    ),
    (
        "bark://localhost",
        {
            # No Device Key specified
            "instance": NotifyBark,
            # Expected notify() response False (because we won't be able
            # to actually notify anything if no device_key was specified
            "notify_response": False,
        },
    ),
    (
        "bark://192.168.0.6:8081/device_key",
        {
            # Everything is okay
            "instance": NotifyBark,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "bark://192.168.0.6:8081/",
        },
    ),
    (
        "bark://user@192.168.0.6:8081/device_key",
        {
            # Everything is okay (test with user)
            "instance": NotifyBark,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "bark://user@192.168.0.6:8081/",
        },
    ),
    (
        "bark://192.168.0.6:8081/device_key/?sound=invalid",
        {
            # bad sound, but we go ahead anyway
            "instance": NotifyBark,
        },
    ),
    (
        "bark://192.168.0.6:8081/device_key/?sound=alarm",
        {
            # alarm.caf sound loaded
            "instance": NotifyBark,
        },
    ),
    (
        "bark://192.168.0.6:8081/device_key/?sound=NOiR.cAf",
        {
            # noir.caf sound loaded
            "instance": NotifyBark,
        },
    ),
    (
        "bark://192.168.0.6:8081/device_key/?badge=100",
        {
            # set badge
            "instance": NotifyBark,
        },
    ),
    (
        "barks://192.168.0.6:8081/device_key/?badge=invalid",
        {
            # set invalid badge
            "instance": NotifyBark,
        },
    ),
    (
        "barks://192.168.0.6:8081/device_key/?badge=-12",
        {
            # set invalid badge
            "instance": NotifyBark,
        },
    ),
    (
        "bark://192.168.0.6:8081/device_key/?category=apprise",
        {
            # set category
            "instance": NotifyBark,
        },
    ),
    (
        "bark://192.168.0.6:8081/device_key/?image=no",
        {
            # do not display image
            "instance": NotifyBark,
        },
    ),
    (
        "bark://192.168.0.6:8081/device_key/?group=apprise",
        {
            # set group
            "instance": NotifyBark,
        },
    ),
    (
        "bark://192.168.0.6:8081/device_key/?level=invalid",
        {
            # bad level, but we go ahead anyway
            "instance": NotifyBark,
        },
    ),
    (
        "bark://192.168.0.6:8081/?to=device_key",
        {
            # test use of to= argument
            "instance": NotifyBark,
        },
    ),
    (
        "bark://192.168.0.6:8081/device_key/?click=http://localhost",
        {
            # Our click link
            "instance": NotifyBark,
        },
    ),
    (
        "bark://192.168.0.6:8081/device_key/?level=active",
        {
            # active level
            "instance": NotifyBark,
        },
    ),
    (
        "bark://192.168.0.6:8081/device_key/?level=critical",
        {
            # critical level
            "instance": NotifyBark,
        },
    ),
    (
        "bark://192.168.0.6:8081/device_key/?level=critical&volume=10",
        {
            # critical level with volume 10
            "instance": NotifyBark,
        },
    ),
    (
        "bark://192.168.0.6:8081/device_key/?level=critical&volume=invalid",
        {
            # critical level with invalid volume
            "instance": NotifyBark,
        },
    ),
    (
        "bark://192.168.0.6:8081/device_key/?level=critical&volume=11",
        {
            # volume > 10
            "instance": NotifyBark,
        },
    ),
    (
        "bark://192.168.0.6:8081/device_key/?level=critical&volume=-1",
        {
            # volume < 0
            "instance": NotifyBark,
        },
    ),
    (
        "bark://192.168.0.6:8081/device_key/?level=critical&volume=",
        {
            # volume None
            "instance": NotifyBark,
        },
    ),
    (
        "bark://user:pass@192.168.0.5:8086/device_key/device_key2/",
        {
            # Everything is okay
            "instance": NotifyBark,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "bark://user:****@192.168.0.5:8086/",
        },
    ),
    (
        "barks://192.168.0.7/device_key/",
        {
            "instance": NotifyBark,
            # throw a bizarre code forcing us to fail to look it up
            "response": False,
            "requests_response_code": 999,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "barks://192.168.0.7/device_key",
        },
    ),
    (
        "bark://192.168.0.7/device_key",
        {
            "instance": NotifyBark,
            # Throws a series of i/o exceptions with this flag
            # is set and tests that we gracefully handle them
            "test_requests_exceptions": True,
        },
    ),
    (
        "bark://192.168.0.6:8081/device_key/?icon=https://example.com/icon.png",
        {
            # set custom icon
            "instance": NotifyBark,
        },
    ),
    (
        "bark://192.168.0.6:8081/device_key/?icon=https://example.com/icon.png&image=no",
        {
            # set custom icon and disable default image
            "instance": NotifyBark,
        },
    ),
    (
        "bark://192.168.0.6:8081/device_key/?call=1",
        {
            # set call parameter to repeat ringtone
            "instance": NotifyBark,
        },
    ),
    (
        "bark://192.168.0.6:8081/device_key/?call=1&sound=alarm&level=critical",
        {
            # set call parameter with other parameters
            "instance": NotifyBark,
        },
    ),
    (
        "bark://192.168.0.6:8081/device_key/?format=markdown",
        {
            # enable markdown mode via global format parameter
            "instance": NotifyBark,
        },
    ),
    (
        "bark://192.168.0.6:8081/device_key/?format=text",
        {
            # explicitly set text format (default behavior)
            "instance": NotifyBark,
        },
    ),
)


def test_plugin_bark_urls():
    """NotifyBark() Apprise URLs."""

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import base64
from datetime import datetime
import json

from ..common import AWARE_DATE_ISO_FORMAT, NAIVE_DATE_ISO_FORMAT
from ..locale import LazyTranslation


class AppriseJSONEncoder(json.JSONEncoder):
    """A JSON Encoder for handling Apprise internals."""

    def default(self, entry):
        if isinstance(entry, datetime):
            return entry.strftime(
                AWARE_DATE_ISO_FORMAT
                if entry.tzinfo is not None
                else NAIVE_DATE_ISO_FORMAT
            )

        elif isinstance(entry, bytes):
            return base64.b64encode(entry).decode("utf-8")

        elif isinstance(entry, (set, frozenset, tuple)):
            return list(entry)

        elif isinstance(entry, LazyTranslation):
            return str(entry)

        return super().default(entry)



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import json

# Disable logging for a cleaner testing output
import logging
import os
from unittest import mock

from helpers import AppriseURLTester
import requests

from apprise import Apprise, AppriseAttachment, NotifyType
from apprise.plugins.custom_json import NotifyJSON

logging.disable(logging.CRITICAL)

# Attachment Directory
TEST_VAR_DIR = os.path.join(os.path.dirname(__file__), "var")

# Our Testing URLs
apprise_url_tests = (
    (
        "json://:@/",
        {
            "instance": None,
        },
    ),
    (
        "json://",
        {
            "instance": None,
        },
    ),
    (
        "jsons://",
        {
            "instance": None,
        },
    ),
    (
        "json://localhost",
        {
            "instance": NotifyJSON,
        },
    ),
    (
        "json://user@localhost?method=invalid",
        {
            "instance": TypeError,
        },
    ),
    (
        "json://user:pass@localhost",
        {
            "instance": NotifyJSON,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "json://user:****@localhost",
        },
    ),
    (
        "json://user@localhost",
        {
            "instance": NotifyJSON,
        },
    ),
    # Test method variations
    (
        "json://user@localhost?method=put",
        {
            "instance": NotifyJSON,
        },
    ),
    (
        "json://user@localhost?method=get",
        {
            "instance": NotifyJSON,
        },
    ),
    (
        "json://user@localhost?method=post",
        {
            "instance": NotifyJSON,
            "force_debug": True,
        },
    ),
    (
        "json://user@localhost?method=head",
        {
            "instance": NotifyJSON,
        },
    ),
    (
        "json://user@localhost?method=delete",
        {
            "instance": NotifyJSON,
        },
    ),
    (
        "json://user@localhost?method=patch",
        {
            "instance": NotifyJSON,
        },
    ),
    (
        "json://user@localhost?method=update",
        {
            "instance": NotifyJSON,
        },
    ),
    (
        "json://user@localhost?method=options",
        {
            "instance": NotifyJSON,
        },
    ),
    # Continue testing other cases
    (
        "json://localhost:8080",
        {
            "instance": NotifyJSON,
        },
    ),
    (
        "json://user:pass@localhost:8080",
        {
            "instance": NotifyJSON,
        },
    ),
    (
        "jsons://localhost",
        {
            "instance": NotifyJSON,
        },
    ),
    (
        "jsons://user:pass@localhost",
        {
            "instance": NotifyJSON,
        },
    ),
    (
        "jsons://localhost:8080/path/",
        {
            "instance": NotifyJSON,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "jsons://localhost:8080/path/",
        },
    ),
    # Test our GET params
    (
        "json://localhost:8080/path?-ParamA=Value",
        {
            "instance": NotifyJSON,
        },
    ),
    (
        "jsons://user:password@localhost:8080",
        {
            "instance": NotifyJSON,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "jsons://user:****@localhost:8080",
        },
    ),
    # Test our Headers
    (
        "json://localhost:8080/path?+HeaderKey=HeaderValue",
        {
            "instance": NotifyJSON,
        },
    ),
    (
        "json://user:pass@localhost:8081",
        {
            "instance": NotifyJSON,
            # force a failure
            "response": False,
            "requests_response_code": requests.codes.internal_server_error,
        },
    ),
    (
        "json://user:pass@localhost:8082",
        {
            "instance": NotifyJSON,
            # throw a bizarre code forcing us to fail to look it up
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "json://user:pass@localhost:8083",
        {
            "instance": NotifyJSON,
            # Throws a series of i/o exceptions with this flag
            # is set and tests that we gracefully handle them
            "test_requests_exceptions": True,
        },
    ),
)


def test_plugin_custom_json_urls():
    """NotifyJSON() Apprise URLs."""

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@mock.patch("requests.request")
def test_plugin_custom_json_edge_cases(mock_request):
    """NotifyJSON() Edge Cases."""

    # Prepare our response
    response = requests.Request()
    response.status_code = requests.codes.ok

    # Prepare Mock
    mock_request.return_value = response

    # This string also tests that type is set to nothing
    results = NotifyJSON.parse_url(
        "json://localhost:8080/command?"
        ":message=msg&:test=value&method=GET"
        "&:type="
    )

    assert isinstance(results, dict)
    assert results["user"] is None
    assert results["password"] is None
    assert results["port"] == 8080
    assert results["host"] == "localhost"
    assert results["fullpath"] == "/command"
    assert results["path"] == "/"
    assert results["query"] == "command"
    assert results["schema"] == "json"
    assert results["url"] == "json://localhost:8080/command"
    assert isinstance(results["qsd:"], dict)
    assert results["qsd:"]["message"] == "msg"
    # empty special mapping
    assert results["qsd:"]["type"] == ""

    instance = NotifyJSON(**results)
    assert isinstance(instance, NotifyJSON)

    response = instance.send(title="title", body="body")
    assert response is True
    assert mock_request.call_count == 1

    details = mock_request.call_args_list[0]
    assert details[0][0] == "GET"
    assert details[0][1] == "http://localhost:8080/command"
    assert "title" in details[1]["data"]
    dataset = json.loads(details[1]["data"])
    assert dataset["title"] == "title"
    assert "message" not in dataset
    # Ensure NotifyType does not leak as an Enum string in JSON
    assert "NotifyType." not in details[1]["data"]
    assert "msg" in dataset
    # type was set to nothing which implies it should be removed
    assert "type" not in dataset
    # message over-ride was provided; the body is now in `msg` and not
    # `message`
    assert dataset["msg"] == "body"

    assert "test" in dataset
    assert dataset["test"] == "value"

    assert instance.url(privacy=False).startswith(
        "json://localhost:8080/command?"
    )

    # Generate a new URL based on our last and verify key values are the same
    new_results = NotifyJSON.parse_url(instance.url(safe=False))
    for k in (
        "user",
        "password",
        "port",
        "host",
        "fullpath",
        "path",
        "query",
        "schema",
        "url",
        "method",
    ):
        assert new_results[k] == results[k]

    mock_request.reset_mock()

    results = NotifyJSON.parse_url(
        "json://localhost:8080/command?:message=msg&method=GET"
    )
    instance = NotifyJSON(**results)
    assert instance.send(title="title", body="body") is True
    details = mock_request.call_args_list[0]
    assert details[0][0] == "GET"
    dataset = json.loads(details[1]["data"])

    assert dataset["type"] == NotifyType.INFO.value
    assert "NotifyType." not in details[1]["data"]

    mock_request.reset_mock()

    # Verify UPDATE method is forwarded correctly
    results = NotifyJSON.parse_url("json://localhost?method=UPDATE")
    instance = NotifyJSON(**results)
    assert instance.send(title="title", body="body") is True
    assert mock_request.call_count == 1
    details = mock_request.call_args_list[0]
    assert details[0][0] == "UPDATE"
    assert details[0][1] == "http://localhost"

    mock_request.reset_mock()

    # Verify OPTIONS method is forwarded correctly
    results = NotifyJSON.parse_url("json://localhost?method=OPTIONS")
    instance = NotifyJSON(**results)
    assert instance.send(title="title", body="body") is True
    assert mock_request.call_count == 1
    details = mock_request.call_args_list[0]
    assert details[0][0] == "OPTIONS"
    assert details[0][1] == "http://localhost"


@mock.patch("requests.request")
def test_notify_json_plugin_attachments(mock_request):
    """NotifyJSON() Attachments."""

    okay_response = requests.Request()
    okay_response.status_code = requests.codes.ok
    okay_response.content = ""

    # Assign our mock object our return value
    mock_request.return_value = okay_response

    obj = Apprise.instantiate("json://localhost.localdomain/")
    assert isinstance(obj, NotifyJSON)

    # Test Valid Attachment
    path = os.path.join(TEST_VAR_DIR, "apprise-test.gif")
    attach = AppriseAttachment(path)
    assert (
        obj.notify(
            body="body",
            title="title",
            notify_type=NotifyType.INFO,
            attach=attach,
        )
        is True
    )

    # Test invalid attachment
    path = os.path.join(TEST_VAR_DIR, "/invalid/path/to/an/invalid/file.jpg")
    assert (
        obj.notify(
            body="body",
            title="title",
            notify_type=NotifyType.INFO,
            attach=path,
        )
        is False
    )

    # Test Valid Attachment (load 3)
    path = (
        os.path.join(TEST_VAR_DIR, "apprise-test.gif"),
        os.path.join(TEST_VAR_DIR, "apprise-test.gif"),
        os.path.join(TEST_VAR_DIR, "apprise-test.gif"),
    )
    attach = AppriseAttachment(path)

    # Return our good configuration
    mock_request.side_effect = None
    mock_request.return_value = okay_response
    with mock.patch("builtins.open", side_effect=OSError()):
        # We can't send the message we can't open the attachment for reading
        assert (
            obj.notify(
                body="body",
                title="title",
                notify_type=NotifyType.INFO,
                attach=attach,
            )
            is False
        )

    # test the handling of our batch modes
    obj = Apprise.instantiate("json://no-reply@example.com/")
    assert isinstance(obj, NotifyJSON)

    # Now send an attachment normally without issues
    mock_request.reset_mock()
    assert (
        obj.notify(
            body="body",
            title="title",
            notify_type=NotifyType.INFO,
            attach=attach,
        )
        is True
    )
    assert mock_request.call_count == 1


# Based on incomming webhook details defined here:
# https://kb.synology.com/en-au/DSM/help/Chat/chat_integration
@mock.patch("requests.request")
def test_plugin_custom_form_for_synology(mock_request):
    """NotifyJSON() Synology Chat Test Case."""

    # Prepare our response
    response = requests.Request()
    response.status_code = requests.codes.ok

    # Prepare Mock
    mock_request.return_value = response

    # This is rather confusing, it may be easier to leverage the
    # synology:// and synologys:// plugins instead, but this is just to prove
    # that the same message can be sent using the json:// plugin.

    results = NotifyJSON.parse_url(
        "jsons://localhost:8081/webapi/entry.cgi?"
        "-api=SYNO.Chat.External&-method=incoming&-version=2&-token=abc123"
        "&:message=text&:version=&:type=&:title=&:attachments"
        "&:file_url=https://i.redd.it/my2t4d2fx0u31.jpg"
    )

    assert isinstance(results, dict)
    assert results["user"] is None
    assert results["password"] is None
    assert results["port"] == 8081
    assert results["host"] == "localhost"
    assert results["fullpath"] == "/webapi/entry.cgi"
    assert results["path"] == "/webapi/"
    assert results["query"] == "entry.cgi"
    assert results["schema"] == "jsons"
    assert results["url"] == "jsons://localhost:8081/webapi/entry.cgi"
    assert isinstance(results["qsd:"], dict)
    # Header Entries
    assert results["qsd-"]["api"] == "SYNO.Chat.External"
    assert results["qsd-"]["method"] == "incoming"
    assert results["qsd-"]["version"] == "2"
    assert results["qsd-"]["token"] == "abc123"

    instance = NotifyJSON(**results)
    assert isinstance(instance, NotifyJSON)

    response = instance.send(title="title", body="body")
    assert response is True
    assert mock_request.call_count == 1

    details = mock_request.call_args_list[0]
    assert details[0][0] == "POST"
    assert details[0][1] == "https://localhost:8081/webapi/entry.cgi"

    params = details[1]["params"]
    assert params.get("api") == "SYNO.Chat.External"
    assert params.get("method") == "incoming"
    assert params.get("version") == "2"
    assert params.get("token") == "abc123"

    payload = json.loads(details[1]["data"])
    assert "version" not in payload
    assert "title" not in payload
    assert "message" not in payload
    assert "type" not in payload
    assert payload.get("text") == "body"
    assert payload.get("file_url") == "https://i.redd.it/my2t4d2fx0u31.jpg"



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from json import dumps

# Disable logging for a cleaner testing output
import logging
import os
from random import choice
import re
from string import ascii_uppercase as str_alpha, digits as str_num
from unittest import mock

import requests

from apprise import (
    Apprise,
    AppriseAsset,
    AppriseAttachment,
    NotifyBase,
    NotifyType,
    PersistentStoreMode,
)
from apprise.common import OverflowMode

logging.disable(logging.CRITICAL)


class AppriseURLTester:
    # Some exception handling we'll use
    req_exceptions = (
        requests.ConnectionError(0, "requests.ConnectionError() not handled"),
        requests.RequestException(
            0, "requests.RequestException() not handled"
        ),
        requests.HTTPError(0, "requests.HTTPError() not handled"),
        requests.ReadTimeout(0, "requests.ReadTimeout() not handled"),
        requests.TooManyRedirects(
            0, "requests.TooManyRedirects() not handled"
        ),
    )

    # Attachment Testing Directory
    __test_var_dir = os.path.join(
        os.path.dirname(os.path.dirname(__file__)), "var"
    )

    # Our URLs we'll test against
    __tests = []

    # Define how many characters exist per line
    row = 80

    # Some variables we use to control the data we work with
    body_len = 1024
    title_len = 1024

    def __init__(self, tests=None, *args, **kwargs):
        """Our initialization."""
        # Create a large body and title with random data
        self.body = "".join(
            choice(str_alpha + str_num + " ") for _ in range(self.body_len)
        )
        self.body = "\r\n".join(
            [
                self.body[i : i + self.row]
                for i in range(0, len(self.body), self.row)
            ]
        )

        # Create our title using random data
        self.title = "".join(
            choice(str_alpha + str_num) for _ in range(self.title_len)
        )

        if tests:
            self.__tests = tests

    def add(self, url, meta):
        """Adds a test suite to our object."""
        self.__tests.append(
            {
                "url": url,
                "meta": meta,
            }
        )

    def run_all(self, tmpdir=None):
        """Run all of our tests."""
        # iterate over our dictionary and test it out
        for url, meta in self.__tests:
            self.run(url, meta, tmpdir)

    @mock.patch("requests.get")
    @mock.patch("requests.post")
    @mock.patch("requests.request")
    def run(self, url, meta, tmpdir, mock_request, mock_post, mock_get):
        """Run a specific test."""

        if meta is False:
            # Prepare a default structure to make life easy
            meta = {
                "instance": TypeError,
            }

        # Our expected instance
        instance = meta.get("instance")

        # Our expected server objects
        self_ = meta.get("self")

        # Our expected privacy url
        # Don't set this if don't need to check it's value
        privacy_url = meta.get("privacy_url")

        # Our regular expression
        url_matches = meta.get("url_matches")

        # Detect our storage path (used to set persistent storage
        # mode
        storage_path = (
            tmpdir
            if tmpdir and isinstance(tmpdir, str) and os.path.isdir(tmpdir)
            else None
        )

        # Our storage mode to set
        storage_mode = meta.get(
            "storage_mode",
            (
                PersistentStoreMode.MEMORY
                if not storage_path
                else PersistentStoreMode.AUTO
            ),
        )

        # Debug Mode
        pdb = meta.get("pdb", False)

        # Whether or not we should include an image with our request; unless
        # otherwise specified, we assume that images are to be included
        include_image = meta.get("include_image", True)
        if include_image:
            # a default asset
            asset = AppriseAsset(
                storage_mode=storage_mode,
                storage_path=storage_path,
            )

        else:
            # Disable images
            asset = AppriseAsset(
                image_path_mask=False,
                image_url_mask=False,
                storage_mode=storage_mode,
                storage_path=storage_path,
            )
            asset.image_url_logo = None

        # Mock our request object
        robj = mock.Mock()
        robj.content = ""
        mock_get.return_value = robj
        mock_post.return_value = robj
        mock_request.return_value = robj

        if pdb:
            # Makes it easier to debug with this peice of code
            # just add `pdb': True to the call that is failing
            import pdb

            pdb.set_trace()

        try:
            # We can now instantiate our object:
            obj = Apprise.instantiate(
                url, asset=asset, suppress_exceptions=False
            )

        except Exception as e:
            # Handle our exception
            if instance is None:
                raise e

            if not isinstance(e, instance):
                raise e

            # We're okay if we get here
            return

        if obj is None:
            if instance is not None:
                # We're done (assuming this is what we were
                # expecting)
                raise AssertionError()
            # We're done because we got the results we expected
            return

        if instance is None:
            # Expected None but didn't get it
            raise AssertionError()

        if not isinstance(obj, instance):
            raise AssertionError()

        if isinstance(obj, NotifyBase):
            # Ensure we are not performing any type of thorttling
            obj.request_rate_per_sec = 0

            # We loaded okay; now lets make sure we can reverse
            # this url
            assert isinstance(obj.url(), str) is True

            # Test that we support a url identifier
            url_id = obj.url_id()

            # It can be either disabled or a string; nothing else
            assert isinstance(url_id, str) or (
                url_id is None and obj.url_identifier is False
            )

            # Verify we can acquire a target count as an integer
            assert isinstance(len(obj), int)

            # Test url() with privacy=True
            assert isinstance(obj.url(privacy=True), str) is True

            # Some Simple Invalid Instance Testing
            assert instance.parse_url(None) is None
            assert instance.parse_url(object) is None
            assert instance.parse_url(42) is None

            # Assess that our privacy url is as expected
            if privacy_url and not obj.url(privacy=True).startswith(
                privacy_url
            ):
                raise AssertionError(
                    f"URL: {url} Privacy URL:"
                    f" '{obj.url(privacy=True)[: len(privacy_url)]}' !="
                    f" expected '{privacy_url}'"
                )

            # Assess that our URL matches a set regex
            if url_matches and not re.search(url_matches, obj.url()):
                raise AssertionError(
                    f"URL: {url} generated an reloadable "
                    f"url() of {obj.url()} that does not match "
                    f"'{url_matches}'"
                )

            # Instantiate the exact same object again using the URL
            # from the one that was already created properly
            obj_cmp = Apprise.instantiate(obj.url())

            if not isinstance(obj_cmp, NotifyBase):
                raise AssertionError(
                    f"URL: {url} generated an un-reloadable "
                    f"url() of {obj.url()}"
                )

            # Our new object should produce the same url identifier
            elif obj.url_identifier != obj_cmp.url_identifier:
                raise AssertionError(
                    f"URL: {url} URL Identifier: "
                    f"'{obj_cmp.url_identifier}' != expected"
                    f" '{obj.url_identifier}'"
                )

            # Back our check up
            if obj.url_id() != obj_cmp.url_id():
                raise AssertionError(
                    f"URL: {url} URL ID(): '{obj_cmp.url_id()}' != expected"
                    f" '{obj.url_id()}'"
                )

            # Verify there is no change from the old and the new
            if len(obj) != len(obj_cmp):
                raise AssertionError(
                    f"URL: {url} generated an reloadable "
                    f"url() of {obj.url()} produced target miscount "
                    f"{len(obj)} != {len(obj_cmp)}"
                )

            # Tidy our object
            del obj_cmp
            del instance

        if self_:
            # Iterate over our expected entries inside of our
            # object
            for key, val in self_.items():
                # Test that our object has the desired key
                assert hasattr(obj, key) is True
                assert getattr(obj, key) == val

        try:
            self.__notify(url, obj, meta, asset)

        except AssertionError:
            # Don't mess with these entries
            raise

        # Tidy our object and allow any possible defined destructors to
        # be executed.
        del obj

    @mock.patch("requests.get")
    @mock.patch("requests.post")
    @mock.patch("requests.head")
    @mock.patch("requests.put")
    @mock.patch("requests.delete")
    @mock.patch("requests.patch")
    @mock.patch("requests.request")
    def __notify(
        self,
        url,
        obj,
        meta,
        asset,
        mock_request,
        mock_patch,
        mock_del,
        mock_put,
        mock_head,
        mock_post,
        mock_get,
    ):
        """Perform notification testing against object specified."""
        #
        # Prepare our options
        #

        # Allow notification type override, otherwise default to INFO
        notify_type = meta.get("notify_type", NotifyType.INFO)

        # Whether or not we're testing exceptions or not
        test_requests_exceptions = meta.get("test_requests_exceptions", False)

        # Our expected Query response (True, False, or exception type)
        response = meta.get("response", True)

        # Our expected Notify response (True or False)
        notify_response = meta.get("notify_response", response)

        # Our expected Notify Attachment response (True or False)
        attach_response = meta.get("attach_response", notify_response)

        # Test attachments
        # Don't set this if don't need to check it's value
        check_attachments = meta.get("check_attachments", True)

        # Debug/Trace Monitoring
        force_debug = meta.get("force_debug", False)

        # Allow us to force the server response code to be something other then
        # the defaults
        requests_response_code = meta.get(
            "requests_response_code",
            requests.codes.ok if response else requests.codes.not_found,
        )

        # Allow us to force the server response text to be something other then
        # the defaults
        requests_response_text = meta.get("requests_response_text")
        requests_response_content = None

        if isinstance(requests_response_text, str):
            requests_response_content = requests_response_text.encode("utf-8")

        elif isinstance(requests_response_text, bytes):
            requests_response_content = requests_response_text
            requests_response_text = requests_response_text.decode("utf-8")

        elif not isinstance(requests_response_text, str):
            # Convert to string
            requests_response_text = dumps(requests_response_text)
            requests_response_content = requests_response_text.encode("utf-8")

        else:
            requests_response_content = ""
            requests_response_text = ""

        # A request
        robj = mock.Mock()
        robj.content = ""
        robj.text = ""
        mock_get.return_value = robj
        mock_post.return_value = robj
        mock_head.return_value = robj
        mock_patch.return_value = robj
        mock_del.return_value = robj
        mock_put.return_value = robj
        mock_request.return_value = robj

        if test_requests_exceptions is False:
            # Handle our default response
            mock_put.return_value.status_code = requests_response_code
            mock_head.return_value.status_code = requests_response_code
            mock_del.return_value.status_code = requests_response_code
            mock_post.return_value.status_code = requests_response_code
            mock_get.return_value.status_code = requests_response_code
            mock_patch.return_value.status_code = requests_response_code
            mock_request.return_value.status_code = requests_response_code

            # Handle our default text response
            mock_get.return_value.content = requests_response_content
            mock_post.return_value.content = requests_response_content
            mock_del.return_value.content = requests_response_content
            mock_put.return_value.content = requests_response_content
            mock_head.return_value.content = requests_response_content
            mock_patch.return_value.content = requests_response_content
            mock_request.return_value.content = requests_response_content

            mock_get.return_value.text = requests_response_text
            mock_post.return_value.text = requests_response_text
            mock_put.return_value.text = requests_response_text
            mock_del.return_value.text = requests_response_text
            mock_head.return_value.text = requests_response_text
            mock_patch.return_value.text = requests_response_text
            mock_request.return_value.text = requests_response_text

            # Ensure there is no side effect set
            mock_post.side_effect = None
            mock_del.side_effect = None
            mock_put.side_effect = None
            mock_head.side_effect = None
            mock_get.side_effect = None
            mock_patch.side_effect = None
            mock_request.side_effect = None

        else:
            # Handle exception testing; first we turn the boolean flag
            # into a list of exceptions
            test_requests_exceptions = self.req_exceptions

        try:
            if test_requests_exceptions is False:
                # Verify we can acquire a target count as an integer
                targets = len(obj)

                # check that we're as expected
                resp = obj.notify(
                    body=self.body, title=self.title, notify_type=notify_type
                )
                if resp != notify_response:
                    raise AssertionError(
                        f"notify() call; notify_response={resp} "
                        f"(expected {notify_response}) on {url}"
                    )

                if notify_response:
                    # If we successfully got a response, there must have been
                    # at least 1 target present
                    assert targets > 0

                # check that this doesn't change using different overflow
                # methods
                assert (
                    obj.notify(
                        body=self.body,
                        title=self.title,
                        notify_type=notify_type,
                        overflow=OverflowMode.UPSTREAM,
                    )
                    == notify_response
                )
                assert (
                    obj.notify(
                        body=self.body,
                        title=self.title,
                        notify_type=notify_type,
                        overflow=OverflowMode.TRUNCATE,
                    )
                    == notify_response
                )
                assert (
                    obj.notify(
                        body=self.body,
                        title=self.title,
                        notify_type=notify_type,
                        overflow=OverflowMode.SPLIT,
                    )
                    == notify_response
                )

                #
                # Handle varations of the Asset Object missing fields
                #

                # First make a backup
                app_id = asset.app_id
                app_desc = asset.app_desc

                # now clear records
                asset.app_id = None
                asset.app_desc = None

                # Notify should still work
                assert (
                    obj.notify(
                        body=self.body,
                        title=self.title,
                        notify_type=notify_type,
                    )
                    == notify_response
                )

                # App ID only
                asset.app_id = app_id
                asset.app_desc = None

                # Notify should still work
                assert (
                    obj.notify(
                        body=self.body,
                        title=self.title,
                        notify_type=notify_type,
                    )
                    == notify_response
                )

                # App Desc only
                asset.app_id = None
                asset.app_desc = app_desc

                if force_debug:
                    # Enable access to areas otherwise locked away by the log
                    # level
                    original_is_enabled_for = obj.logger.isEnabledFor
                    original_debug = obj.logger.debug
                    try:
                        # Force code paths guarded by isEnabledFor(DEBUG)
                        obj.logger.isEnabledFor = mock.Mock(return_value=True)

                        # Prevent actual logging emission (avoids pytest
                        # capture closed stream)
                        obj.logger.debug = mock.Mock()

                        assert (
                            obj.notify(
                                body=self.body,
                                title=self.title,
                                notify_type=notify_type,
                            )
                            == notify_response
                        )

                    finally:
                        # Restore
                        obj.logger.isEnabledFor = original_is_enabled_for
                        obj.logger.debug = original_debug
                else:
                    assert (
                        obj.notify(
                            body=self.body,
                            title=self.title,
                            notify_type=notify_type,
                        )
                        == notify_response
                    )

                # Restore
                asset.app_id = app_id
                asset.app_desc = app_desc

                if check_attachments:
                    # Test single attachment support; even if the service
                    # doesn't support attachments, it should still
                    # gracefully ignore the data
                    attach = os.path.join(
                        self.__test_var_dir, "apprise-test.gif"
                    )
                    assert (
                        obj.notify(
                            body=self.body,
                            title=self.title,
                            notify_type=notify_type,
                            attach=attach,
                        )
                        == attach_response
                    )

                    # Same results should apply to a list of attachments
                    attach = AppriseAttachment(
                        (
                            os.path.join(
                                self.__test_var_dir, "apprise-test.gif"
                            ),
                            os.path.join(
                                self.__test_var_dir, "apprise-test.png"
                            ),
                            os.path.join(
                                self.__test_var_dir, "apprise-test.jpeg"
                            ),
                        )
                    )

                    assert (
                        obj.notify(
                            body=self.body,
                            title=self.title,
                            notify_type=notify_type,
                            attach=attach,
                        )
                        == attach_response
                    )

                    if obj.attachment_support:
                        #
                        # Services that support attachments should support
                        # sending a attachment (or more) without a body or
                        # title specified:
                        #
                        assert (
                            obj.notify(
                                body=None,
                                title=None,
                                notify_type=notify_type,
                                attach=attach,
                            )
                            == attach_response
                        )

                        # Turn off attachment support on the notifications
                        # that support it so we can test that any logic we
                        # have ot test against this flag is ran
                        obj.attachment_support = False

                        #
                        # Notifications should still transmit as normal if
                        # Attachment support is flipped off
                        #
                        assert (
                            obj.notify(
                                body=self.body,
                                title=self.title,
                                notify_type=notify_type,
                                attach=attach,
                            )
                            == notify_response
                        )

                        #
                        # We should not be able to send a message without a
                        # body or title in this circumstance
                        #
                        assert (
                            obj.notify(
                                body=None,
                                title=None,
                                notify_type=notify_type,
                                attach=attach,
                            )
                            is False
                        )

                        # Toggle Back
                        obj.attachment_support = True

                    else:  # No Attachment support
                        #
                        # We should not be able to send a message without a
                        # body or title in this circumstance
                        #
                        assert (
                            obj.notify(
                                body=None,
                                title=None,
                                notify_type=notify_type,
                                attach=attach,
                            )
                            is False
                        )
            else:
                for exception in self.req_exceptions:
                    mock_post.side_effect = exception
                    mock_head.side_effect = exception
                    mock_del.side_effect = exception
                    mock_put.side_effect = exception
                    mock_get.side_effect = exception
                    mock_patch.side_effect = exception
                    mock_request.side_effect = exception

                    try:
                        assert (
                            obj.notify(
                                body=self.body,
                                title=self.title,
                                notify_type=notify_type,
                            )
                            is False
                        )

                    except AssertionError:
                        # Don't mess with these entries
                        raise

                    except Exception:
                        # We can't handle this exception type
                        raise

        except AssertionError:
            # Don't mess with these entries
            raise

        except Exception as e:
            # Check that we were expecting this exception to happen
            try:
                if not isinstance(e, response):
                    raise e

            except TypeError:
                raise e  # noqa: B904 - intentional re-raising in test helper

        #
        # Do the test again but without a title defined
        #
        try:
            if test_requests_exceptions is False:
                # check that we're as expected
                assert (
                    obj.notify(body="body", notify_type=notify_type)
                    == notify_response
                )

            else:
                for exception in self.req_exceptions:
                    mock_post.side_effect = exception
                    mock_del.side_effect = exception
                    mock_put.side_effect = exception
                    mock_head.side_effect = exception
                    mock_get.side_effect = exception
                    mock_patch.side_effect = exception
                    mock_request.side_effect = exception

                    try:
                        assert (
                            obj.notify(body=self.body, notify_type=notify_type)
                            is False
                        )

                    except AssertionError:
                        # Don't mess with these entries
                        raise

                    except Exception:
                        # We can't handle this exception type
                        raise

        except AssertionError:
            # Don't mess with these entries
            raise

        except Exception as e:
            # Check that we were expecting this exception to happen
            if not isinstance(e, response):
                raise e

        return True



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Disable logging for a cleaner testing output
import logging
import os
from unittest import mock

from helpers import AppriseURLTester
import pytest
import requests

from apprise import Apprise, AppriseAttachment, NotifyType
from apprise.plugins.brevo import NotifyBrevo

logging.disable(logging.CRITICAL)

# Attachment Directory
TEST_VAR_DIR = os.path.join(os.path.dirname(__file__), "var")

# a test UUID we can use
UUID4 = "8b799edf-6f98-4d3a-9be7-2862fb4e5752"

# Our Testing URLs
apprise_url_tests = (
    (
        "brevo://",
        {
            "instance": None,
        },
    ),
    (
        "brevo://:@/",
        {
            "instance": None,
        },
    ),
    (
        "brevo://abcd",
        {
            # Just an broken email (no api key or email)
            "instance": None,
        },
    ),
    (
        "brevo://abcd@host",
        {
            # Just an Email specified, no API Key
            "instance": None,
        },
    ),
    (
        "brevo://invalid-api-key+*-d:user@example.com",
        {
            # An invalid API Key
            "instance": TypeError,
        },
    ),
    (
        ("brevo://abcd:user@example.com/newuser@example.com?reply=%20!"),
        {
            # An invalid Reply-To address
            "instance": TypeError,
        },
    ),
    (
        "brevo://abcd:user@example.com?format=text",
        {
            # No To/Target Address(es) specified; so we sub in the same From
            # address
            "instance": NotifyBrevo,
        },
    ),
    (
        (
            "brevo://abcd:user@example.com/newuser@example.com"
            "?reply=user@example.ca"
        ),
        {
            # A good email
            "instance": NotifyBrevo,
            "force_debug": True,
        },
    ),
    (
        "brevo://abcd:user@example.com/bademailaddress",
        {
            # won't be able to send email
            "instance": NotifyBrevo,
            "notify_response": False,
        },
    ),
    (
        (
            "brevo://abcd:user@example.com/newuser@example.com"
            "?bcc=l2g@nuxref.com"
        ),
        {
            # A good email with Blind Carbon Copy
            "instance": NotifyBrevo,
        },
    ),
    (
        (
            "brevo://abcd:user@example.com/newuser@example.com"
            "?cc=l2g@nuxref.com"
        ),
        {
            # A good email with Carbon Copy
            "instance": NotifyBrevo,
        },
    ),
    (
        (
            "brevo://abcd:user@example.com/newuser@example.com"
            "?to=l2g@nuxref.com"
        ),
        {
            # A good email with Carbon Copy
            "instance": NotifyBrevo,
        },
    ),
    (
        "brevo://abcd:user@example.ca/newuser@example.ca",
        {
            "instance": NotifyBrevo,
            # force a failure
            "response": False,
            "requests_response_code": requests.codes.internal_server_error,
        },
    ),
    (
        "brevo://abcd:user@example.uk/newuser@example.uk",
        {
            "instance": NotifyBrevo,
            # throw a bizzare code forcing us to fail to look it up
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "brevo://abcd:user@example.au/newuser@example.au",
        {
            "instance": NotifyBrevo,
            # Throws a series of i/o exceptions with this flag
            # is set and tests that we gracfully handle them
            "test_requests_exceptions": True,
        },
    ),
)


def test_plugin_brevo_urls():
    """NotifyBrevo() Apprise URLs."""

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@mock.patch("requests.get")
@mock.patch("requests.post")
def test_plugin_brevo_edge_cases(mock_post, mock_get):
    """NotifyBrevo() Edge Cases."""

    # no apikey
    with pytest.raises(TypeError):
        NotifyBrevo(apikey=None, from_email="user@example.com")

    # invalid from email
    with pytest.raises(TypeError):
        NotifyBrevo(apikey="abcd", from_email="!invalid")

    # no email
    with pytest.raises(TypeError):
        NotifyBrevo(apikey="abcd", from_email=None)

    # Invalid To email address
    NotifyBrevo(
        apikey="abcd", from_email="user@example.com", targets="!invalid"
    )

    # Test invalid bcc/cc entries mixed with good ones
    assert isinstance(
        NotifyBrevo(
            apikey="abcd",
            from_email="l2g@example.com",
            bcc=("abc@def.com", "!invalid"),
            cc=("abc@test.org", "!invalid"),
        ),
        NotifyBrevo,
    )


@mock.patch("requests.post")
def test_plugin_brevo_attachments(mock_post):
    """NotifyBrevo() Attachments."""

    request = mock.Mock()
    request.status_code = requests.codes.ok

    # Prepare Mock
    mock_post.return_value = request

    path = os.path.join(TEST_VAR_DIR, "apprise-test.gif")
    attach = AppriseAttachment(path)
    obj = Apprise.instantiate("brevo://abcd:user@example.com")
    assert isinstance(obj, NotifyBrevo)
    assert (
        obj.notify(
            body="body",
            title="title",
            notify_type=NotifyType.INFO,
            attach=attach,
        )
        is True
    )

    mock_post.reset_mock()

    # Try again in a use case where we can't access the file
    with mock.patch("os.path.isfile", return_value=False):
        assert (
            obj.notify(
                body="body",
                title="title",
                notify_type=NotifyType.INFO,
                attach=attach,
            )
            is False
        )

    # Try again in a use case where we can't access the file
    with mock.patch("builtins.open", side_effect=OSError):
        assert (
            obj.notify(
                body="body",
                title="title",
                notify_type=NotifyType.INFO,
                attach=attach,
            )
            is False
        )



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Disable logging for a cleaner testing output
import logging
import os
from unittest import mock

from helpers import AppriseURLTester
import requests

from apprise import Apprise, AppriseAttachment, NotifyType
from apprise.plugins.smtp2go import NotifySMTP2Go

logging.disable(logging.CRITICAL)

# Attachment Directory
TEST_VAR_DIR = os.path.join(os.path.dirname(__file__), "var")

# Our Testing URLs
apprise_url_tests = (
    (
        "smtp2go://",
        {
            "instance": TypeError,
        },
    ),
    (
        "smtp2go://:@/",
        {
            "instance": TypeError,
        },
    ),
    # No Token specified
    (
        "smtp2go://user@localhost.localdomain",
        {
            "instance": TypeError,
        },
    ),
    # Token is valid, but no user name specified
    (
        "smtp2go://localhost.localdomain/{}-{}-{}".format(
            "a" * 32, "b" * 8, "c" * 8
        ),
        {
            "instance": TypeError,
        },
    ),
    # Invalid from email address
    (
        'smtp2go://"@localhost.localdomain/{}-{}-{}'.format(
            "a" * 32, "b" * 8, "c" * 8
        ),
        {
            "instance": TypeError,
        },
    ),
    # No To email address, but everything else is valid
    (
        "smtp2go://user@localhost.localdomain/{}-{}-{}".format(
            "a" * 32, "b" * 8, "c" * 8
        ),
        {
            "instance": NotifySMTP2Go,
        },
    ),
    (
        "smtp2go://user@localhost.localdomain/{}-{}-{}?format=markdown".format(
            "a" * 32, "b" * 8, "c" * 8
        ),
        {
            "instance": NotifySMTP2Go,
        },
    ),
    (
        "smtp2go://user@localhost.localdomain/{}-{}-{}?format=html".format(
            "a" * 32, "b" * 8, "c" * 8
        ),
        {
            "instance": NotifySMTP2Go,
        },
    ),
    (
        "smtp2go://user@localhost.localdomain/{}-{}-{}?format=text".format(
            "a" * 32, "b" * 8, "c" * 8
        ),
        {
            "instance": NotifySMTP2Go,
        },
    ),
    # headers
    (
        "smtp2go://user@localhost.localdomain/{}-{}-{}"
        "?+X-Customer-Campaign-ID=Apprise".format("a" * 32, "b" * 8, "c" * 8),
        {
            "instance": NotifySMTP2Go,
            "force_debug": True,
        },
    ),
    # bcc and cc
    (
        "smtp2go://user@localhost.localdomain/{}-{}-{}"
        "?bcc=user@example.com&cc=user2@example.com".format(
            "a" * 32, "b" * 8, "c" * 8
        ),
        {
            "instance": NotifySMTP2Go,
        },
    ),
    # One To Email address
    (
        "smtp2go://user@localhost.localdomain/{}-{}-{}/test@example.com".format(
            "a" * 32, "b" * 8, "c" * 8
        ),
        {
            "instance": NotifySMTP2Go,
        },
    ),
    (
        "smtp2go://user@localhost.localdomain/"
        "{}-{}-{}?to=test@example.com".format("a" * 32, "b" * 8, "c" * 8),
        {"instance": NotifySMTP2Go},
    ),
    # One To Email address, a from name specified too
    (
        "smtp2go://user@localhost.localdomain/{}-{}-{}/"
        'test@example.com?name="Frodo"'.format("a" * 32, "b" * 8, "c" * 8),
        {"instance": NotifySMTP2Go},
    ),
    # Invalid 'To' Email address
    (
        "smtp2go://user@localhost.localdomain/{}-{}-{}/invalid".format(
            "a" * 32, "b" * 8, "c" * 8
        ),
        {
            "instance": NotifySMTP2Go,
            # Expected notify() response
            "notify_response": False,
        },
    ),
    # Multiple 'To', 'Cc', and 'Bcc' addresses (with invalid ones)
    (
        "smtp2go://user@example.com/{}-{}-{}/{}?bcc={}&cc={}".format(
            "a" * 32,
            "b" * 8,
            "c" * 8,
            "/".join(
                ("user1@example.com", "invalid", "User2:user2@example.com")
            ),
            ",".join(("user3@example.com", "i@v", "User1:user1@example.com")),
            ",".join(("user4@example.com", "g@r@b", "Da:user5@example.com")),
        ),
        {
            "instance": NotifySMTP2Go,
        },
    ),
    (
        "smtp2go://user@localhost.localdomain/{}-{}-{}".format(
            "a" * 32, "b" * 8, "c" * 8
        ),
        {
            "instance": NotifySMTP2Go,
            # force a failure
            "response": False,
            "requests_response_code": requests.codes.internal_server_error,
        },
    ),
    (
        "smtp2go://user@localhost.localdomain/{}-{}-{}".format(
            "a" * 32, "b" * 8, "c" * 8
        ),
        {
            "instance": NotifySMTP2Go,
            # throw a bizarre code forcing us to fail to look it up
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "smtp2go://user@localhost.localdomain/{}-{}-{}".format(
            "a" * 32, "b" * 8, "c" * 8
        ),
        {
            "instance": NotifySMTP2Go,
            # Throws a series of i/o exceptions with this flag
            # is set and tests that we gracefully handle them
            "test_requests_exceptions": True,
        },
    ),
)


def test_plugin_smtp2go_urls():
    """NotifySMTP2Go() Apprise URLs."""

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@mock.patch("requests.post")
def test_plugin_smtp2go_attachments(mock_post):
    """NotifySMTP2Go() Attachments."""

    okay_response = requests.Request()
    okay_response.status_code = requests.codes.ok
    okay_response.content = ""

    # Assign our mock object our return value
    mock_post.return_value = okay_response

    # API Key
    apikey = "abc123"

    obj = Apprise.instantiate(f"smtp2go://user@localhost.localdomain/{apikey}")
    assert isinstance(obj, NotifySMTP2Go)

    # Test Valid Attachment
    path = os.path.join(TEST_VAR_DIR, "apprise-test.gif")
    attach = AppriseAttachment(path)
    assert (
        obj.notify(
            body="body",
            title="title",
            notify_type=NotifyType.INFO,
            attach=attach,
        )
        is True
    )

    # Test invalid attachment
    path = os.path.join(TEST_VAR_DIR, "/invalid/path/to/an/invalid/file.jpg")
    assert (
        obj.notify(
            body="body",
            title="title",
            notify_type=NotifyType.INFO,
            attach=path,
        )
        is False
    )

    mock_post.return_value = None
    mock_post.side_effect = OSError()
    # We can't send the message if we can't read the attachment
    assert (
        obj.notify(
            body="body",
            title="title",
            notify_type=NotifyType.INFO,
            attach=attach,
        )
        is False
    )

    # Test Valid Attachment (load 3)
    path = (
        os.path.join(TEST_VAR_DIR, "apprise-test.gif"),
        os.path.join(TEST_VAR_DIR, "apprise-test.gif"),
        os.path.join(TEST_VAR_DIR, "apprise-test.gif"),
    )
    attach = AppriseAttachment(path)

    # Return our good configuration
    mock_post.side_effect = None
    mock_post.return_value = okay_response
    with mock.patch("builtins.open", side_effect=OSError()):
        # We can't send the message we can't open the attachment for reading
        assert (
            obj.notify(
                body="body",
                title="title",
                notify_type=NotifyType.INFO,
                attach=attach,
            )
            is False
        )

    # test the handling of our batch modes
    obj = Apprise.instantiate(
        f"smtp2go://no-reply@example.com/{apikey}/"
        "user1@example.com/user2@example.com?batch=yes"
    )
    assert isinstance(obj, NotifySMTP2Go)

    # objects will be combined into a single post in batch mode
    assert len(obj) == 1

    # Force our batch to break into separate messages
    obj.default_batch_size = 1

    # We'll send 2 messages now
    assert len(obj) == 2

    mock_post.reset_mock()

    assert (
        obj.notify(
            body="body",
            title="title",
            notify_type=NotifyType.INFO,
            attach=attach,
        )
        is True
    )
    assert mock_post.call_count == 2

    # single batch
    mock_post.reset_mock()
    # We'll send 1 message
    obj.default_batch_size = 2

    assert (
        obj.notify(
            body="body",
            title="title",
            notify_type=NotifyType.INFO,
            attach=attach,
        )
        is True
    )
    assert mock_post.call_count == 1



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Unit tests for Apprise IRC plugin and helper modules."""

from __future__ import annotations

import logging
import time
from typing import Any, Optional
from unittest import mock

import pytest

from apprise.plugins.irc import NotifyIRC
from apprise.plugins.irc.client import IRCClient
from apprise.plugins.irc.protocol import (
    IRCAuthMode,
    IRCMessage,
    extract_welcome_nick,
    is_ping,
    normalise_channel,
    parse_irc_line,
    ping_payload,
)
from apprise.plugins.irc.state import (
    IRCActionKind,
    IRCContext,
    IRCState,
    IRCStateMachine,
)
from apprise.utils.socket import AppriseSocketError

# Disable logging for a cleaner testing output
logging.disable(logging.CRITICAL)


class _DummyTransport:
    """A lightweight SocketTransport stand-in for IRCClient tests."""

    def __init__(self) -> None:
        self.connected = False
        self.closed = False
        self.writes: list[bytes] = []
        self.reads: list[bytes] = []

    def connect(self) -> None:
        self.connected = True

    def close(self) -> None:
        self.closed = True

    def write(self, payload: bytes, *, flush: bool, timeout: float) -> None:
        self.writes.append(payload)

    def read(
        self,
        max_bytes: int,
        *,
        blocking: bool,
        timeout: Optional[float],
    ) -> bytes:
        if not self.reads:
            return b""
        chunk = self.reads.pop(0)
        return chunk[:max_bytes]


def test_plugin_irc_init_targets() -> None:
    """NotifyIRC parses targets."""
    with mock.patch.object(NotifyIRC, "apply_irc_defaults"):
        n = NotifyIRC(
            host="irc.example.com",
            targets=["#chan", "%23chan2:key", "@bob", "%40alice", "  bad# "],
        )

    assert "chan" in n.channels
    assert n.channels["chan"] is None
    assert "chan2" in n.channels
    assert n.channels["chan2"] == "key"
    assert "bob" in n.users
    assert "alice" in n.users


def test_plugin_irc_modes() -> None:
    """NotifyIRC auth mode tests."""
    with (
        mock.patch.object(NotifyIRC, "apply_irc_defaults"),
        pytest.raises(TypeError),
    ):
        NotifyIRC(host="irc.example.com", targets=["#c"], mode="invalid")

    with mock.patch.object(NotifyIRC, "apply_irc_defaults"):
        # Case Insensitive
        result = NotifyIRC(
            host="irc.example.com", targets=["#c"], mode="NICKServ"
        )
        assert "mode=nickserv" in result.url()

        # Case Insensitive
        result = NotifyIRC(
            host="irc.example.com", targets=["#c"], mode="server"
        )
        assert "mode=" not in result.url()


def test_plugin_irc_defaults_port_noop() -> None:
    """NotifyIRC defaults are not applied when port is explicit."""
    n = NotifyIRC(host="irc.efnet.org", targets=["#c"], port=7000)
    n.apply_irc_defaults()
    assert n.port == 7000


def test_plugin_irc_defaults_template_match() -> None:
    """NotifyIRC defaults apply for known networks."""
    n = NotifyIRC(host="irc.synirc.net", targets=["#c"])
    assert n.secure is True
    assert n.port == 6697
    assert n.auth_mode == IRCAuthMode.NICKSERV


def test_plugin_irc_defaults_template_none() -> None:
    """NotifyIRC defaults do not force unknown networks."""
    n = NotifyIRC(host="irc.unknown.tld", targets=["#c"])
    assert n.secure is False


def test_plugin_irc_send_no_targets() -> None:
    """NotifyIRC send() rejects empty targets."""
    n = NotifyIRC(host="irc.example.com", targets=[])
    assert n.send("body") is False


def test_plugin_irc_send_ok() -> None:
    """NotifyIRC send() valid path."""
    n = NotifyIRC(
        host="irc.example.com",
        targets=["#chan:key", "@bob"],
        user="me",
        password="pw",
    )
    n.join = True

    client = mock.Mock(spec=IRCClient)
    client.nickname = "me"

    with mock.patch("apprise.plugins.irc.base.IRCClient", return_value=client):
        assert n.send("body", title="title") is True

    client.connect.assert_called_once()
    client.register.assert_called_once()
    client.join.assert_called_once()
    client.privmsg.assert_any_call(
        target="#chan",
        message="title body",
        timeout=mock.ANY,
    )
    client.privmsg.assert_any_call(
        target="bob",
        message="title body",
        timeout=mock.ANY,
    )
    client.quit.assert_called_once()
    client.close.assert_called_once()


def test_plugin_irc_send_no_join() -> None:
    """NotifyIRC avoids JOIN when join is false and no key is provided."""
    n = NotifyIRC(host="irc.example.com", targets=["#chan", "@bob"])
    n.join = False

    client = mock.Mock(spec=IRCClient)
    client.nickname = "x"

    with mock.patch("apprise.plugins.irc.base.IRCClient", return_value=client):
        assert n.send("body") is True

    client.join.assert_not_called()
    client.privmsg.assert_any_call(
        target="#chan",
        message="body",
        timeout=mock.ANY,
    )


def test_plugin_irc_send_error() -> None:
    """NotifyIRC returns False on connect errors."""
    n = NotifyIRC(host="irc.example.com", targets=["#chan"])
    client = mock.Mock(spec=IRCClient)
    client.connect.side_effect = AppriseSocketError("boom")

    with mock.patch("apprise.plugins.irc.base.IRCClient", return_value=client):
        assert n.send("body") is False

    client.close.assert_called_once()


def test_plugin_irc_url_id() -> None:
    """NotifyIRC url_identifier."""
    n = NotifyIRC(
        host="irc.example.com", targets=["#c"], user="me", password="pw"
    )
    assert n.url_identifier == ("irc", "irc.example.com", "me", "pw")

    n.secure = True
    assert n.url_identifier[0] == "ircs"


def test_plugin_irc_url_format() -> None:
    """NotifyIRC url() basic rendering and privacy."""
    n = NotifyIRC(
        host="irc.example.com",
        targets=["#chan:key", "@bob"],
        user="me",
        password="pw",
        secure=False,
        port=IRCClient.default_insecure_port,
        nick="nick",
        name="Real Name",
        join=False,
        mode=IRCAuthMode.NICKSERV,
    )

    url = n.url(privacy=False)
    assert url.startswith("irc://me:pw@irc.example.com/")
    assert "#chan" in url
    assert "@bob" in url

    # Mode only appears when auth_mode is not SERVER.
    if n.auth_mode != IRCAuthMode.SERVER:
        assert "mode=nickserv" in url
    else:
        assert "mode=" not in url

    private = n.url(privacy=True)
    assert "pw" not in private
    assert "key" not in private

    n = NotifyIRC(
        host="irc.example.com",
        targets=["#chan", "@bob"],
        user="user2",
        secure=True,
        port=IRCClient.default_insecure_port,
        join=False,
        mode=IRCAuthMode.NICKSERV,
    )

    url = n.url(privacy=False)
    assert url.startswith("ircs://user2@irc.example.com:6667/")
    assert "#chan" in url
    assert "@bob" in url

    # Mode only appears when auth_mode is not SERVER.
    if n.auth_mode != IRCAuthMode.SERVER:
        assert "mode=nickserv" in url
    else:
        assert "mode=" not in url

    private = n.url(privacy=True)
    assert "pw" not in private
    assert "key" not in private


def test_plugin_irc_parse_url() -> None:
    """NotifyIRC parse_url() behaviour with host= query."""

    with mock.patch(
        "apprise.plugins.irc.base.NotifyBase.parse_url",
        return_value=None,
    ):
        results = NotifyIRC.parse_url("irc://%@@")
        assert results is None

    # host= indicates the URL host is a target, it does not override host.
    results = NotifyIRC.parse_url("irc://%23chan?host=irc.example.com")
    assert results is not None
    assert results["host"] == "%23chan"
    assert "#chan" in results["targets"]

    results = NotifyIRC.parse_url(
        "irc://irc.example.com/%23a/@b?to=%23c,@d"
        "&join=no&name=Z&nick=N&mode=none"
    )
    assert results is not None
    assert results["host"] == "irc.example.com"
    assert "#a" in results["targets"]
    assert "@b" in results["targets"]
    assert "#c" in results["targets"]
    assert "@d" in results["targets"]
    assert results["join"] is False
    assert results["name"] == "Z"
    assert results["nick"] == "N"
    assert results["mode"] == "none"


def test_plugin_irc_protocol() -> None:
    """Protocol helpers."""
    msg = parse_irc_line(":srv 001 nick :welcome")
    assert msg.numeric == 1
    assert extract_welcome_nick(msg) == "nick"

    ping = parse_irc_line("PING :abc")
    assert is_ping(ping) is True
    assert ping_payload(ping) == "abc"

    assert normalise_channel("chan") == "#chan"
    assert normalise_channel("#chan") == "#chan"
    assert normalise_channel("    ") == ""


def test_plugin_irc_state_machine() -> None:
    """State machine basics."""
    ctx = IRCContext(
        desired_nick="n", accepted_nick="n", fullname="f", password="pw"
    )
    sm = IRCStateMachine(ctx)

    acts = sm.start_registration()
    assert any(
        a.kind == IRCActionKind.SEND and a.line and a.line.startswith("PASS ")
        for a in acts
    )

    sm.on_message(parse_irc_line(":srv 001 nick :welcome"))
    assert sm.ctx.registered is True
    assert sm.ctx.accepted_nick == "nick"

    ctx2 = IRCContext(
        desired_nick="n", accepted_nick="n", fullname="f", password="pw"
    )
    sm2 = IRCStateMachine(ctx2)
    sm2.start_registration()
    fail = sm2.on_message(parse_irc_line(":srv 464 n :bad pass"))
    assert fail and fail[0].kind == IRCActionKind.FAIL

    ctx3 = IRCContext(desired_nick="n", accepted_nick="n", fullname="f")
    sm3 = IRCStateMachine(ctx3)
    sm3.start_registration()
    sm3.on_message(parse_irc_line(":srv 001 n :welcome"))
    sm3.request_join("#c", key=None)
    sm3.on_message(parse_irc_line(":srv 366 n #c :End of /NAMES list."))
    assert "#c" in sm3.ctx.joined


def test_plugin_irc_client_nick_handling() -> None:
    """IRCClient nickname handling."""
    nick0 = IRCClient.nick_generation(prefix="Apprise", length=9, collision=0)
    assert len(nick0) == 9

    nick1 = IRCClient.nick_generation(prefix="Apprise", length=9, collision=12)
    assert len(nick1) == 9
    assert nick1[-1].isdigit()

    client = IRCClient(
        host="irc.example.com",
        nickname="nick",
        fullname="full",
        nick_generator=None,
    )
    with pytest.raises(AppriseSocketError):
        client._nickname_collision_handler(prefix="x")

    def gen(prefix: str, length: int, collision: int) -> str:
        return f"{prefix}{collision}"

    client = IRCClient(
        host="irc.example.com",
        nickname="nick",
        fullname="full",
        nick_generator=gen,
    )
    client._nick_collision = int(client.nickname_collision_max)
    with pytest.raises(AppriseSocketError):
        client._nickname_collision_handler(prefix="x")


def test_plugin_irc_client_handshake() -> None:
    """IRCClient handshake error path."""

    def gen(prefix: str, length: int, collision: int) -> str:
        return "newnick"

    client = IRCClient(
        host="irc.example.com",
        nickname="nick",
        fullname="full",
        nick_generator=gen,
    )
    client.transport = _DummyTransport()

    # Put the state machine into REGISTERING so 464 triggers FAIL.
    client.sm.start_registration()

    lines = [
        "PING :abc",
        ":srv 433 nick :in use",
        ":srv 464 nick :bad pass",
        "",
    ]

    def _read(deadline: float) -> Optional[str]:
        return lines.pop(0)

    writes: list[str] = []

    def _write(line: Any, deadline: float) -> None:
        if isinstance(line, bytes):
            line = line.decode("utf-8", errors="replace")
        writes.append(str(line))

    with (
        mock.patch.object(client, "_read", side_effect=_read),
        mock.patch.object(client, "_write", side_effect=_write),
        pytest.raises(AppriseSocketError),
    ):
        client._handshake(deadline=time.monotonic() + 1.0, prefix="ap")

    assert any(w.startswith("PONG") for w in writes)
    assert any(w.startswith("NICK ") for w in writes)


def test_plugin_irc_client_register() -> None:
    """IRCClient register timeout and success."""
    client = IRCClient(
        host="irc.example.com",
        nickname="nick",
        fullname="full",
        password="pw",
        auth_mode=IRCAuthMode.NICKSERV,
    )
    client.transport = _DummyTransport()

    with (
        mock.patch("time.monotonic", side_effect=[0.0, 1.0]),
        pytest.raises(TimeoutError),
    ):
        client.register(timeout=0.01, prefix="ap")

    client2 = IRCClient(
        host="irc.example.com",
        nickname="nick",
        fullname="full",
        password="pw",
        auth_mode=IRCAuthMode.NICKSERV,
    )
    client2.transport = _DummyTransport()

    def _handshake(deadline: float, prefix: str) -> None:
        client2.sm.ctx.registered = True

    with (
        mock.patch.object(client2, "_flush"),
        mock.patch.object(client2, "_handshake", side_effect=_handshake),
        mock.patch.object(client2, "identify") as m_identify,
    ):
        client2.register(timeout=0.1, prefix="ap")
        m_identify.assert_called_once()


def test_plugin_irc_client_identify() -> None:
    """IRCClient identify early exits and send."""
    client = IRCClient(
        host="irc.example.com",
        nickname="nick",
        fullname="full",
        password=None,
        auth_mode=IRCAuthMode.NICKSERV,
    )
    client.transport = _DummyTransport()
    client.identify(timeout=0.1)

    client2 = IRCClient(
        host="irc.example.com",
        nickname="nick",
        fullname="full",
        password="pw",
        auth_mode=IRCAuthMode.SERVER,
    )
    client2.transport = _DummyTransport()
    client2.identify(timeout=0.1)

    client3 = IRCClient(
        host="irc.example.com",
        nickname="nick",
        fullname="full",
        password="pw",
        auth_mode=IRCAuthMode.NICKSERV,
    )
    client3.transport = _DummyTransport()
    with (
        mock.patch.object(client3, "_flush") as m_flush,
        mock.patch.object(client3, "_handshake") as m_hs,
    ):
        client3.identify(timeout=0.1)
        m_flush.assert_called_once()
        m_hs.assert_called_once()


def test_plugin_irc_client_join() -> None:
    """IRCClient join timeout path."""
    client = IRCClient(
        host="irc.example.com",
        nickname="nick",
        fullname="full",
    )
    client.transport = _DummyTransport()

    # join() calls monotonic frequently; use a callable so we never exhaust
    # a finite side_effect list.
    calls = {"n": 0}

    def _mono() -> float:
        calls["n"] += 1
        # Stay below deadline long enough to exercise the loop, then exceed it.
        return 0.0 if calls["n"] < 2000 else 1.0

    with (
        mock.patch.object(client, "_flush") as m_flush,
        mock.patch.object(client, "_handshake") as m_hs,
        mock.patch("time.monotonic", side_effect=_mono),
    ):
        client.join(channel="#chan", timeout=0.01, prefix="ap", key=None)
        assert m_flush.called
        assert m_hs.called

    with mock.patch.object(client, "_write") as m_write:
        client.quit(message="bye", timeout=0.1)
        assert m_write.called


def test_plugin_irc_protocol_parse_blank_line() -> None:
    """Protocol parse blank input."""
    msg = parse_irc_line("   ")
    assert msg.command == ""
    assert msg.params == ()
    assert msg.trailing is None


def test_plugin_irc_protocol_parse_trailing_only() -> None:
    """Protocol parse trailing only."""
    msg = parse_irc_line(" :hello")
    assert msg.command == ""
    assert msg.params == ()
    assert msg.trailing == "hello"


def test_plugin_irc_message() -> None:
    """IRCMessage testing"""

    message = IRCMessage("raw", None, "not-a-number", ("a", "b"), None)
    assert message.numeric is None


def test_plugin_irc_protocol_ping_payload_from_params() -> None:
    """Ping payload from params."""
    msg = parse_irc_line("PING abc")
    assert msg.trailing is None
    assert msg.params == ("abc",)
    assert ping_payload(msg) == "abc"


def test_plugin_irc_protocol_extract_welcome_nick_non_welcome() -> None:
    """Welcome nick ignores non-001."""
    msg = parse_irc_line(":srv 002 nick :Your host is")
    assert msg.numeric == 2
    assert extract_welcome_nick(msg) is None


def test_plugin_irc_protocol_normalise_channel_empty() -> None:
    """Normalise empty channel."""
    assert normalise_channel("") == ""


def test_plugin_irc_client_props_and_io() -> None:
    """Client basics."""
    c = IRCClient(host="h", nickname="n", fullname="f")
    c.transport = mock.Mock()
    c.sm.ctx.accepted_nick = "nick"
    assert c.nickname == "nick"
    c.connect()
    c.close()
    c.transport.connect.assert_called_once()
    c.transport.close.assert_called_once()


def test_plugin_irc_client_write_timeout() -> None:
    """Write timeout."""
    c = IRCClient(host="h", nickname="n", fullname="f")
    c.transport = mock.Mock()
    with (
        mock.patch("time.monotonic", return_value=10.0),
        pytest.raises(TimeoutError),
    ):
        c._write("X", deadline=9.0)


def test_plugin_irc_client_write_bytes_and_flush() -> None:
    """Write bytes and flush queue."""
    c = IRCClient(host="h", nickname="n", fullname="f")
    c.transport = mock.Mock()

    with mock.patch("time.monotonic", return_value=0.0):
        c._write(b"RAW", deadline=1.0)
        c.transport.write.assert_called_with(b"RAW", flush=True, timeout=1.0)

    c.transport.write.reset_mock()
    c._queue("A")
    c._queue("B")
    with mock.patch.object(c, "_write") as m_write:
        c._flush(deadline=1.0)
        assert m_write.call_count == 2
        assert len(c._out_queue) == 0


def test_plugin_irc_client_read_buffer_and_timeout() -> None:
    """Read buffer and timeout."""
    c = IRCClient(host="h", nickname="n", fullname="f")
    c.transport = mock.Mock()

    c._inbuf = bytearray(b"hello\r\n")
    assert c._read(deadline=time.monotonic() + 1.0) == "hello"

    c._inbuf = bytearray()
    with mock.patch("time.monotonic", return_value=10.0):
        assert c._read(deadline=9.0) is None


def test_plugin_irc_client_read_transport_paths() -> None:
    """Read from transport."""
    c = IRCClient(host="h", nickname="n", fullname="f")
    c.transport = mock.Mock()

    with mock.patch("time.monotonic", return_value=0.0):
        c.transport.read.return_value = b""
        assert c._read(deadline=1.0) is None

    c = IRCClient(host="h", nickname="n", fullname="f")
    c.transport = mock.Mock()
    with mock.patch("time.monotonic", side_effect=[0.0, 0.0, 0.0]):
        c.transport.read.side_effect = [b"he", b"llo\n"]
        assert c._read(deadline=1.0) == "hello"


def test_plugin_irc_client_tick() -> None:
    """Tick timing."""
    c = IRCClient(host="h", nickname="n", fullname="f")
    with mock.patch("time.monotonic", return_value=10.0):
        assert c._tick(deadline=9.0) == pytest.approx(9.0)

    with mock.patch("time.monotonic", return_value=0.0):
        # remaining=0.5 -> 0.0 + 0.5
        assert c._tick(deadline=0.5) == pytest.approx(0.5)


def test_plugin_irc_client_handshake_paths() -> None:
    """Handshake paths."""
    c = IRCClient(host="h", nickname="n", fullname="f")
    c.transport = mock.Mock()

    with mock.patch.object(c, "_read", return_value=None):
        c._handshake(deadline=time.monotonic() + 1.0, prefix="")

    lines = ["PING :abc", ""]
    with (
        mock.patch.object(
            c, "_read", side_effect=lambda deadline: lines.pop(0)
        ),
        mock.patch.object(c, "_write") as m_write,
    ):
        c._handshake(deadline=time.monotonic() + 1.0, prefix="")
        m_write.assert_any_call("PONG :abc", deadline=mock.ANY)

    lines = [":srv 433 n :in use", ""]
    with (
        mock.patch.object(
            c, "_read", side_effect=lambda deadline: lines.pop(0)
        ),
        mock.patch.object(
            c, "_nickname_collision_handler", return_value="new"
        ),
        mock.patch.object(c, "_write") as m_write,
    ):
        c._handshake(deadline=time.monotonic() + 1.0, prefix="ap")
        m_write.assert_any_call("NICK new", deadline=mock.ANY)

    # FAIL action raises
    from apprise.plugins.irc.state import IRCAction, IRCActionKind

    lines = [":srv 464 n :bad pass", ""]
    with (
        mock.patch.object(
            c, "_read", side_effect=lambda deadline: lines.pop(0)
        ),
        mock.patch.object(
            c.sm,
            "on_message",
            return_value=[IRCAction(IRCActionKind.FAIL, reason="boom")],
        ),
        pytest.raises(AppriseSocketError),
    ):
        c._handshake(deadline=time.monotonic() + 1.0, prefix="")

    # SEND action writes line
    lines = [":srv 001 n :welcome", ""]
    with (
        mock.patch.object(
            c, "_read", side_effect=lambda deadline: lines.pop(0)
        ),
        mock.patch.object(
            c.sm,
            "on_message",
            return_value=[IRCAction(IRCActionKind.SEND, line="NICK x")],
        ),
        mock.patch.object(c, "_write") as m_write,
    ):
        c._handshake(deadline=time.monotonic() + 1.0, prefix="")
        m_write.assert_any_call("NICK x", deadline=mock.ANY)


def test_plugin_irc_client_register_auth_modes() -> None:
    """Register auth modes."""
    c = IRCClient(
        host="h",
        nickname="n",
        fullname="f",
        password="pw",
        auth_mode=IRCAuthMode.NICKSERV,
    )
    c.transport = mock.Mock()

    # Force loop to exit by setting registered in handshake
    def _hs(deadline: float, prefix: str) -> None:
        c.sm.ctx.registered = True

    with (
        mock.patch.object(c, "_flush"),
        mock.patch.object(c, "_handshake", side_effect=_hs),
        mock.patch.object(c, "identify") as m_ident,
        mock.patch("time.monotonic", return_value=0.0),
        mock.patch("time.time", side_effect=[0.0, 0.1, 0.2]),
    ):
        c.register(timeout=1.0, prefix="ap")
        # password cleared before registration when not SERVER
        assert c.sm.ctx.password is None
        m_ident.assert_called_once()

    c2 = IRCClient(
        host="h",
        nickname="n",
        fullname="f",
        password="pw",
        auth_mode=IRCAuthMode.SERVER,
    )
    c2.transport = mock.Mock()

    def _hs2(deadline: float, prefix: str) -> None:
        c2.sm.ctx.registered = True

    with (
        mock.patch.object(c2, "_flush"),
        mock.patch.object(c2, "_handshake", side_effect=_hs2),
        mock.patch.object(c2, "identify") as m_ident2,
        mock.patch("time.monotonic", return_value=0.0),
        mock.patch("time.time", side_effect=[0.0, 0.1, 0.2]),
    ):
        c2.register(timeout=1.0, prefix="ap")
        # identify not called for SERVER
        m_ident2.assert_not_called()


def test_plugin_irc_client_join_privmsg_identify_quit() -> None:
    """Join and message helpers."""
    c = IRCClient(
        host="h",
        nickname="n",
        fullname="f",
        password="pw",
        auth_mode=IRCAuthMode.NICKSERV,
    )
    c.transport = mock.Mock()

    calls = {"n": 0}

    def _mono() -> float:
        calls["n"] += 1
        # Return 0.0 long enough to enter the loop, then jump past deadline
        # to ensure join() exits.
        return 0.0 if calls["n"] < 5000 else 1.0

    with (
        mock.patch.object(c, "_flush") as m_flush,
        mock.patch.object(c, "_handshake") as m_hs,
        mock.patch("time.monotonic", side_effect=_mono),
    ):
        # join never confirmed -> debug path
        with mock.patch("apprise.plugins.irc.client.logger.debug") as m_dbg:
            c.join(channel="#c", timeout=0.01, prefix="ap", key=None)
            m_dbg.assert_called()

        assert m_flush.called
        assert m_hs.called

    # privmsg() and identify() also call monotonic; keep them in a separate
    # context with a simple stable time.
    with (
        mock.patch.object(c, "_flush") as m_flush2,
        mock.patch.object(c, "_handshake") as m_hs2,
        mock.patch("time.monotonic", return_value=0.0),
    ):
        c.privmsg(target="#c", message="m", timeout=0.1)
        c.identify(timeout=0.1)
        assert m_flush2.called
        assert m_hs2.called

    # identify exits early
    c2 = IRCClient(
        host="h",
        nickname="n",
        fullname="f",
        password=None,
        auth_mode=IRCAuthMode.NICKSERV,
    )
    c2.transport = mock.Mock()
    with mock.patch.object(c2, "_flush") as m_flush3:
        c2.identify(timeout=0.1)
        m_flush3.assert_not_called()

    c3 = IRCClient(
        host="h",
        nickname="n",
        fullname="f",
        password="pw",
        auth_mode=IRCAuthMode.SERVER,
    )
    c3.transport = mock.Mock()
    with mock.patch.object(c3, "_flush") as m_flush4:
        c3.identify(timeout=0.1)
        m_flush4.assert_not_called()

    # quit queues and flushes
    c4 = IRCClient(host="h", nickname="n", fullname="f")
    c4.transport = mock.Mock()
    with mock.patch.object(c4, "_flush") as m_flush5:
        c4.quit(message="bye", timeout=0.1)
        m_flush5.assert_called_once()


def test_plugin_irc_client_nick_generation_default_length() -> None:
    """Nick generation defaults."""
    nick = IRCClient.nick_generation(prefix="Ap", length=None, collision=0)
    assert len(nick) == IRCClient.nickname_max_length


def test_plugin_irc_client_write_trace() -> None:
    """Write trace logging."""
    c = IRCClient(host="h", nickname="n", fullname="f")
    c.transport = mock.Mock()

    with (
        mock.patch("time.monotonic", return_value=0.0),
        mock.patch(
            "apprise.plugins.irc.client.logger.isEnabledFor", return_value=True
        ),
        mock.patch("apprise.plugins.irc.client.logger.trace") as m_trace,
    ):
        c._write("HELLO", deadline=1.0)
        m_trace.assert_called_once()
        # Ensure we wrote CRLF terminated bytes
        c.transport.write.assert_called_once()
        payload = c.transport.write.call_args.args[0]
        assert payload.endswith(b"\r\n")


def test_plugin_irc_client_handshake_send_without_line() -> None:
    """Handshake ignores empty SEND."""
    c = IRCClient(host="h", nickname="n", fullname="f")
    c.transport = mock.Mock()

    from apprise.plugins.irc.state import IRCAction, IRCActionKind

    # One message then stop the loop
    lines = [":srv 001 n :welcome", ""]
    with (
        mock.patch.object(
            c, "_read", side_effect=lambda deadline: lines.pop(0)
        ),
        mock.patch.object(
            c.sm,
            "on_message",
            return_value=[IRCAction(IRCActionKind.SEND, line="")],
        ),
        mock.patch.object(c, "_write") as m_write,
    ):
        c._handshake(deadline=time.monotonic() + 1.0, prefix="")
        m_write.assert_not_called()


def test_plugin_irc_client_register_queue_ignores_empty_send() -> None:
    """Register ignores empty SEND."""
    c = IRCClient(
        host="h", nickname="n", fullname="f", auth_mode=IRCAuthMode.NONE
    )
    c.transport = mock.Mock()

    from apprise.plugins.irc.state import IRCAction, IRCActionKind

    with (
        mock.patch.object(
            c.sm,
            "start_registration",
            return_value=[
                IRCAction(IRCActionKind.SEND, line=None),
                IRCAction(IRCActionKind.SEND, line=""),
            ],
        ),
        mock.patch.object(c, "_queue") as m_queue,
        mock.patch.object(c, "_flush"),
        mock.patch.object(c, "_handshake"),
        mock.patch("time.monotonic", return_value=0.0),
        mock.patch("time.time", return_value=0.0),
    ):
        # Make it "already registered" to avoid the while loop behaviour
        c.sm.ctx.registered = True
        c.register(timeout=0.1, prefix="ap")
        m_queue.assert_not_called()


def test_plugin_irc_client_join_queue_ignores_empty_send() -> None:
    """Join ignores empty SEND."""
    c = IRCClient(host="h", nickname="n", fullname="f")
    c.transport = mock.Mock()

    from apprise.plugins.irc.state import IRCAction, IRCActionKind

    with (
        mock.patch.object(
            c.sm,
            "request_join",
            return_value=[
                IRCAction(IRCActionKind.SEND, line=None),
                IRCAction(IRCActionKind.SEND, line=""),
            ],
        ),
        mock.patch.object(c, "_queue") as m_queue,
        mock.patch.object(c, "_flush"),
        mock.patch.object(c, "_handshake"),
        mock.patch("time.monotonic", return_value=1.0),
    ):
        # deadline will be 1.0 + timeout, but we also ensure we do not enter
        # loop by setting joined to include channel.
        c.sm.ctx.joined.add("#c")
        c.join(channel="#c", timeout=0.1, prefix="ap", key=None)
        m_queue.assert_not_called()


def test_plugin_irc_client_quit_queue_ignores_empty_send() -> None:
    """Quit ignores empty SEND."""
    c = IRCClient(host="h", nickname="n", fullname="f")
    c.transport = mock.Mock()

    from apprise.plugins.irc.state import IRCAction, IRCActionKind

    with (
        mock.patch.object(
            c.sm,
            "request_quit",
            return_value=[
                IRCAction(IRCActionKind.SEND, line=None),
                IRCAction(IRCActionKind.SEND, line=""),
            ],
        ),
        mock.patch.object(c, "_queue") as m_queue,
        mock.patch.object(c, "_flush") as m_flush,
        mock.patch("time.monotonic", return_value=0.0),
    ):
        c.quit(message="bye", timeout=0.1)
        m_queue.assert_not_called()
        m_flush.assert_called_once()


def test_plugin_irc_send_znc_calls_check_connection() -> None:
    """NotifyIRC ZNC mode triggers a connection check."""
    n = NotifyIRC(
        host="irc.example.com",
        targets=["#chan"],
        user="zncuser",
        password="zncpass",
        mode=IRCAuthMode.ZNC,
        nick="mybot",
    )

    client = mock.Mock(spec=IRCClient)
    client.nickname = "mybot"
    client.check_connection.return_value = True

    with mock.patch("apprise.plugins.irc.base.IRCClient", return_value=client):
        assert n.send("body") is True

    client.connect.assert_called_once()
    client.register.assert_called_once()
    client.check_connection.assert_called_once()
    client.privmsg.assert_called_once()
    client.quit.assert_called_once()
    client.close.assert_called_once()


def test_plugin_irc_send_znc_check_connection_failure() -> None:
    """NotifyIRC ZNC mode fails fast if the bouncer liveness check fails."""
    n = NotifyIRC(
        host="irc.example.com",
        targets=["#chan"],
        user="zncuser",
        password="zncpass",
        mode=IRCAuthMode.ZNC,
        nick="mybot",
    )

    client = mock.Mock(spec=IRCClient)
    client.nickname = "mybot"
    client.check_connection.return_value = False

    with mock.patch("apprise.plugins.irc.base.IRCClient", return_value=client):
        assert n.send("body") is False

    client.connect.assert_called_once()
    client.register.assert_called_once()
    client.check_connection.assert_called_once()
    client.privmsg.assert_not_called()
    client.quit.assert_not_called()
    client.close.assert_called_once()


def test_plugin_irc_send_znc_pass_rewrite() -> None:
    """NotifyIRC ZNC mode passes user:pass to IRCClient password."""
    n = NotifyIRC(
        host="irc.example.com",
        targets=["#chan"],
        user="zncuser",
        password="zncpass",
        mode=IRCAuthMode.ZNC,
        nick="mybot",
    )

    client = mock.Mock(spec=IRCClient)
    client.nickname = "mybot"
    client.check_connection.return_value = True

    with mock.patch(
        "apprise.plugins.irc.base.IRCClient", return_value=client
    ) as m_ctor:
        assert n.send("body") is True

    # Pull the kwargs passed into IRCClient(...)
    kwargs = m_ctor.call_args.kwargs
    assert kwargs["password"] == "zncuser:zncpass"


def test_plugin_irc_client_check_connection_success_on_any_pong() -> None:
    """check_connection() returns True when a PONG is observed."""
    c = IRCClient(host="h", nickname="n", fullname="f")
    c.transport = mock.Mock()

    # 0.0 used to compute deadline, then remain within deadline once, then stop
    times = [0.0, 0.0, 0.0]

    def _mono() -> float:
        return times.pop(0) if times else 10.0

    with (
        mock.patch("time.monotonic", side_effect=_mono),
        mock.patch.object(c, "_write") as m_write,
        mock.patch.object(c, "_read", side_effect=["PONG :server"]) as m_read,
    ):
        assert c.check_connection(timeout=5.0) is True
        m_write.assert_called_once()
        m_read.assert_called()


def test_plugin_irc_client_handles_empty_reads() -> None:
    """check_connection() handles empty reads and returns False at deadline."""
    c = IRCClient(host="h", nickname="n", fullname="f")
    c.transport = mock.Mock()

    # Start at 0.0, enter loop once, then exceed deadline after the empty read.
    times = [0.0, 0.0, 6.0]

    def _mono() -> float:
        return times.pop(0)

    with (
        mock.patch("time.monotonic", side_effect=_mono),
        mock.patch.object(c, "_write") as m_write,
        mock.patch.object(c, "_read", return_value=None),
    ):
        assert c.check_connection(timeout=5.0) is False
        m_write.assert_called_once()


def test_plugin_irc_client_join_queues_join_line_and_completes() -> None:
    """join() queues the JOIN command when request_join yields a SEND line."""
    c = IRCClient(host="h", nickname="n", fullname="f")
    c.transport = mock.Mock()

    def _handshake(deadline: float, prefix: str) -> None:
        # Simulate join confirmation being observed
        c.sm.ctx.joined.add("#c")

    with (
        mock.patch("time.monotonic", return_value=0.0),
        mock.patch.object(c, "_queue") as m_queue,
        mock.patch.object(c, "_flush"),
        mock.patch.object(c, "_handshake", side_effect=_handshake),
    ):
        c.join(channel="#c", timeout=1.0, prefix="ap", key=None)

    # Ensure the queue branch was taken
    assert any(
        call.args
        and isinstance(call.args[0], str)
        and call.args[0].startswith("JOIN ")
        for call in m_queue.call_args_list
    )


def test_plugin_irc_client_join_timeout_logs_debug() -> None:
    """join() emits a debug line when confirmation is not observed."""
    c = IRCClient(host="h", nickname="n", fullname="f")
    c.transport = mock.Mock()

    # timeout=0.0 forces the while-loop to be skipped and triggers debug path
    with (
        mock.patch("time.monotonic", return_value=0.0),
        mock.patch("apprise.plugins.irc.client.logger.debug") as m_dbg,
        mock.patch.object(c, "_flush"),
        mock.patch.object(c, "_handshake"),
    ):
        c.join(channel="#c", timeout=0.0, prefix="ap", key=None)
        m_dbg.assert_called_once()


def test_plugin_irc_state_machine_joining_numeric_443_adds_channel() -> None:
    """State machine treats numeric 443 as joined confirmation."""
    ctx = IRCContext(desired_nick="n", accepted_nick="n", fullname="f")
    sm = IRCStateMachine(ctx)

    # Enter JOINING state
    sm.request_join("#c")

    assert sm.state == IRCState.JOINING

    # 443 usually indicates already on channel; this should mark as joined.
    sm.on_message(parse_irc_line(":srv 443 n #c :is already on channel"))

    assert "#c" in sm.ctx.joined
    assert sm.state == IRCState.READY


def test_plugin_irc_client_connection_check() -> None:
    """check_connection() sees a non-PONG message."""
    c = IRCClient(host="h", nickname="n", fullname="f")
    c.transport = mock.Mock()

    # monotonic() is used to compute deadline, then loop condition.
    # We provide values such that we enter the loop at least once, then
    # exceed deadline.
    times = [0.0, 0.0, 0.0, 6.0]

    def _mono() -> float:
        return times.pop(0)

    with (
        mock.patch("time.monotonic", side_effect=_mono),
        mock.patch.object(c, "_write") as m_write,
        # First read returns a non-empty line that is not a PONG, then nothing.
        mock.patch.object(c, "_read", side_effect=["PING :abc", None]),
    ):
        assert c.check_connection(timeout=5.0) is False
        m_write.assert_called_once()


def test_plugin_irc_client_join_non_send_actions() -> None:
    """join() ignores non-SEND actions"""
    c = IRCClient(host="h", nickname="n", fullname="f")
    c.transport = mock.Mock()

    from apprise.plugins.irc.state import IRCAction, IRCActionKind

    # Return a NOOP action so: act.kind == SEND is False
    actions = [IRCAction(IRCActionKind.NOOP)]

    # Skip the while loop by making monotonic() already past deadline
    with (
        mock.patch.object(c.sm, "request_join", return_value=actions),
        mock.patch.object(c, "_queue") as m_queue,
        mock.patch.object(c, "_flush") as m_flush,
        mock.patch.object(c, "_handshake") as m_hs,
        mock.patch("time.monotonic", side_effect=[0.0, 1.0]),
        mock.patch("apprise.plugins.irc.client.logger.debug") as m_dbg,
    ):
        c.join(channel="#c", timeout=0.0, prefix="ap", key=None)

        # Nothing queued because our action was not SEND
        m_queue.assert_not_called()

        # While loop is skipped, so these are not called
        m_flush.assert_not_called()
        m_hs.assert_not_called()

        # Not joined, so debug path executes
        m_dbg.assert_called_once()



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

from datetime import datetime, timedelta, timezone
from json import loads

# Disable logging for a cleaner testing output
import logging
import os
from random import choice
from string import ascii_uppercase as str_alpha, digits as str_num
from unittest import mock

from helpers import AppriseURLTester
import pytest
import requests

from apprise import Apprise, AppriseAttachment, NotifyFormat, NotifyType
from apprise.common import OverflowMode
from apprise.plugins.fluxer import NotifyFluxer

logging.disable(logging.CRITICAL)

# Attachment Directory
TEST_VAR_DIR = os.path.join(os.path.dirname(__file__), "var")


def _tokens() -> tuple[str, str]:
    """Return tokens that satisfy Fluxer validation."""

    # webhook_id: digits, >= 10
    # webhook_token: [A-Za-z0-9_-], >= 16
    return ("0" * 10, "B" * 40)


# Our Testing URLs
apprise_url_tests = (
    (
        "fluxer://",
        {
            "instance": TypeError,
        },
    ),
    # An invalid url
    (
        "fluxer://:@/",
        {
            "instance": TypeError,
        },
    ),
    # No webhook_token specified
    (
        "fluxer://%s" % ("0" * 10),
        {
            "instance": TypeError,
        },
    ),
    # Provide both a webhook id and a webhook token
    (
        "fluxer://{}/{}".format(*_tokens()),
        {
            "instance": NotifyFluxer,
            "requests_response_code": requests.codes.no_content,
        },
    ),
    # Provide a temporary username
    (
        "fluxer://l2g@{}/{}".format(*_tokens()),
        {
            "instance": NotifyFluxer,
            "requests_response_code": requests.codes.no_content,
        },
    ),
    # Mode set to private, but no hostname was fluxers; we toggle to cloud mode
    (
        "fluxer://api.fluxer.app/{}/{}?mode=private".format(*_tokens()),
        {
            "instance": NotifyFluxer,
            "requests_response_code": requests.codes.no_content,
        },
    ),
    # test image= field
    (
        "fluxer://{}/{}?format=markdown&footer=Yes&image=Yes&ping=Joe".format(
            *_tokens()
        ),
        {
            "instance": NotifyFluxer,
            "requests_response_code": requests.codes.no_content,
            # don't include an image by default
            "include_image": False,
        },
    ),
    (
        "fluxer://{}/{}?format=markdown&footer=Yes&image=No&fields=no".format(
            *_tokens()
        ),
        {
            "instance": NotifyFluxer,
            "requests_response_code": requests.codes.no_content,
        },
    ),
    (
        "fluxer://jack@{}/{}?format=markdown&footer=Yes&image=Yes".format(
            *_tokens()
        ),
        {
            "instance": NotifyFluxer,
            "requests_response_code": requests.codes.no_content,
        },
    ),
    (
        "fluxer://jack@{}/{}?mode=private&host=example.ca".format(*_tokens()),
        {
            "instance": NotifyFluxer,
            "requests_response_code": requests.codes.no_content,
        },
    ),
    (
        "fluxer://{}/{}?mode=private&host=example.ca&name=jack".format(
            *_tokens()
        ),
        {
            "instance": NotifyFluxer,
            "requests_response_code": requests.codes.no_content,
        },
    ),
    (
        "fluxer://example.ca:123/{}/{}".format(*_tokens()),
        {
            "instance": NotifyFluxer,
            "requests_response_code": requests.codes.no_content,
        },
    ),
    (
        # Invalid Mode
        "fluxer://jack@{}/{}?mode=invalid".format(*_tokens()),
        {
            "instance": TypeError,
        },
    ),
    (
        "https://api.fluxer.app/webhooks/{}/{}".format(*_tokens()),
        {
            # Native URL Support
            "instance": NotifyFluxer,
            "requests_response_code": requests.codes.no_content,
        },
    ),
    (
        "https://api.fluxer.app/v1/webhooks/{}/{}?footer=yes".format(
            *_tokens()
        ),
        {
            # Native URL Support with arguments
            "instance": NotifyFluxer,
            "requests_response_code": requests.codes.no_content,
            "privacy_url": "fluxer://0...0/B...B/",
        },
    ),
    (
        "https://api.fluxer.app/v1/webhooks/{}/{}?footer=yes&botname=joe".format(
            *_tokens()
        ),
        {
            # Native URL Support with arguments
            "instance": NotifyFluxer,
            "requests_response_code": requests.codes.no_content,
            "privacy_url": "fluxer://joe@0...0/B...B/",
        },
    ),
    (
        "fluxer://{}/{}?format=markdown&avatar=No&footer=No".format(
            *_tokens()
        ),
        {
            "instance": NotifyFluxer,
            "requests_response_code": requests.codes.no_content,
        },
    ),
    (
        "fluxer://{}/{}?flags=1".format(*_tokens()),
        {
            "instance": NotifyFluxer,
            "requests_response_code": requests.codes.no_content,
        },
    ),
    (
        "fluxer://{}/{}?flags=-1".format(*_tokens()),
        {
            # invalid flags specified (variation 1)
            "instance": TypeError,
        },
    ),
    (
        "fluxer://{}/{}?flags=invalid".format(*_tokens()),
        {
            # invalid flags specified (variation 2)
            "instance": TypeError,
        },
    ),
    # different format support
    (
        "fluxer://{}/{}?format=markdown".format(*_tokens()),
        {
            "instance": NotifyFluxer,
            "requests_response_code": requests.codes.no_content,
        },
    ),
    # Thread ID (forces markdown mode)
    (
        "fluxer://{}/{}?format=markdown&thread=abc123".format(*_tokens()),
        {
            "instance": NotifyFluxer,
            "requests_response_code": requests.codes.no_content,
        },
    ),
    (
        "fluxer://{}/{}?format=text".format(*_tokens()),
        {
            "instance": NotifyFluxer,
            "requests_response_code": requests.codes.no_content,
        },
    ),
    # Test with href (title link)
    (
        "fluxer://{}/{}?hmarkdown=true&ref=http://localhost".format(
            *_tokens()
        ),
        {
            "instance": NotifyFluxer,
            "requests_response_code": requests.codes.no_content,
        },
    ),
    # Test with url (title link) - Alias of href
    (
        "fluxer://{}/{}?markdown=true&url=http://localhost".format(*_tokens()),
        {
            "instance": NotifyFluxer,
            "requests_response_code": requests.codes.no_content,
        },
    ),
    # Test with avatar URL
    (
        "fluxer://{}/{}?avatar_url=http://localhost/test.jpg".format(
            *_tokens()
        ),
        {
            "instance": NotifyFluxer,
            "requests_response_code": requests.codes.no_content,
        },
    ),
    # Test without image set
    (
        "fluxer://{}/{}".format(*_tokens()),
        {
            "instance": NotifyFluxer,
            "requests_response_code": requests.codes.no_content,
            # don't include an image by default
            "include_image": False,
        },
    ),
    (
        "fluxer://{}/{}".format(*_tokens()),
        {
            "instance": NotifyFluxer,
            # force a failure
            "response": False,
            "requests_response_code": requests.codes.internal_server_error,
        },
    ),
    (
        "fluxer://{}/{}".format(*_tokens()),
        {
            "instance": NotifyFluxer,
            # throw a bizarre code forcing us to fail to look it up
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "fluxer://{}/{}".format(*_tokens()),
        {
            "instance": NotifyFluxer,
            # Throws a series of i/o exceptions with this flag set and tests
            # that we gracefully handle them
            "test_requests_exceptions": True,
        },
    ),
)


def test_plugin_fluxer_urls() -> None:
    """NotifyFluxer() Apprise URLs."""

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@mock.patch("requests.post")
def test_plugin_fluxer_notifications(mock_post: mock.MagicMock) -> None:
    """NotifyFluxer() Notifications/Ping Support."""

    webhook_id, webhook_token = _tokens()

    # Prepare Mock
    mock_post.return_value = requests.Request()
    mock_post.return_value.status_code = requests.codes.ok
    mock_post.return_value.content = b""
    mock_post.return_value.headers = {}

    # Test our header parsing when not lead with a header
    body = """
    # Heading
    @everyone and @admin, wake and meet our new user <@123>; <@&456>"
    """

    results = NotifyFluxer.parse_url(
        f"fluxer://{webhook_id}/{webhook_token}/?format=markdown"
    )

    assert isinstance(results, dict)
    assert results["user"] is None
    assert results["webhook_id"] == webhook_id
    assert results["webhook_token"] == webhook_token
    assert results["schema"] == "fluxer"

    instance = NotifyFluxer(**results)
    assert isinstance(instance, NotifyFluxer)

    assert instance.send(body=body) is True
    assert mock_post.call_count == 1

    details = mock_post.call_args_list[0]
    assert (
        details[0][0]
        == f"https://api.fluxer.app/webhooks/{webhook_id}/{webhook_token}"
    )

    payload = loads(details[1]["data"])
    assert isinstance(payload, dict)

    assert "allow_mentions" in payload
    assert set(payload["allow_mentions"]["users"]) == {"123"}
    assert set(payload["allow_mentions"]["roles"]) == {"456"}
    assert set(payload["allow_mentions"]["parse"]) == {"everyone", "admin"}
    assert payload["content"].startswith("👉 ")

    # Reset our object
    mock_post.reset_mock()

    results = NotifyFluxer.parse_url(
        f"fluxer://{webhook_id}/{webhook_token}/?format=text"
    )

    instance = NotifyFluxer(**results)
    assert isinstance(instance, NotifyFluxer)

    assert instance.send(body=body) is True
    payload = loads(mock_post.call_args_list[0][1]["data"])

    # text mode does not parse mentions from body
    assert "allow_mentions" not in payload

    # Reset our object
    mock_post.reset_mock()

    body = " "
    results = NotifyFluxer.parse_url(
        # & -> %26 for role otherwise & separates our URL from further parsing
        f"fluxer://{webhook_id}/{webhook_token}/?ping=@joe,<@321>,<@%26654>"
    )
    instance = NotifyFluxer(**results)
    assert isinstance(instance, NotifyFluxer)

    assert instance.send(body=body) is True
    payload = loads(mock_post.call_args_list[0][1]["data"])

    assert "allow_mentions" in payload
    assert set(payload["allow_mentions"]["users"]) == {"321"}
    assert set(payload["allow_mentions"]["roles"]) == {"654"}
    assert set(payload["allow_mentions"]["parse"]) == {"joe"}
    assert "<@321>" in payload["content"]
    assert "<@&654>" in payload["content"]
    assert "@joe" in payload["content"]

    # Reset our object
    mock_post.reset_mock()

    # Test our body in text mode, with ping=set
    body = """
    # Heading
    @everyone and @admin, wake and meet our new user <@123>; <@&456>"
    """

    results = NotifyFluxer.parse_url(
        # & -> %26 for role otherwise & separates our URL from further parsing
        f"fluxer://{webhook_id}/{webhook_token}/?ping=@joe,<@321>,<@%26654>"
        "&format=text"
    )

    instance = NotifyFluxer(**results)
    assert isinstance(instance, NotifyFluxer)

    assert instance.send(body=body) is True
    payload = loads(mock_post.call_args_list[0][1]["data"])

    # Payload only includes elements on ping= line with text mode
    assert "allow_mentions" in payload
    assert set(payload["allow_mentions"]["users"]) == {"321"}
    assert set(payload["allow_mentions"]["roles"]) == {"654"}
    assert set(payload["allow_mentions"]["parse"]) == {"joe"}


@mock.patch("requests.post")
@mock.patch("time.sleep")
def test_plugin_fluxer_429(
    mock_sleep: mock.MagicMock,
    mock_post: mock.MagicMock,
) -> None:
    """
    NotifyFluxer() 429 handling

    Focus: Fluxer-specific HTTP 429 + Retry-After handling (recursive retry),
    including header parsing edge cases and retry exhaustion.
    """

    # Prevent throttling side effects
    mock_sleep.return_value = True

    webhook_id, webhook_token = _tokens()

    # Basic construction checks (keep these, they match plugin validation)
    with pytest.raises(TypeError):
        NotifyFluxer(webhook_id=None, webhook_token=webhook_token)
    with pytest.raises(TypeError):
        NotifyFluxer(webhook_id="  ", webhook_token=webhook_token)

    with pytest.raises(TypeError):
        NotifyFluxer(webhook_id=webhook_id, webhook_token=None)
    with pytest.raises(TypeError):
        NotifyFluxer(webhook_id=webhook_id, webhook_token="   ")

    obj = NotifyFluxer(
        webhook_id=webhook_id,
        webhook_token=webhook_token,
        footer=True,
        include_image=True,
    )

    # url() must always return a string
    assert isinstance(obj.url(), str)

    # Helper to build responses
    def _resp(code: int, headers: dict[str, str] | None = None) -> mock.Mock:
        r = mock.Mock()
        r.status_code = code
        r.content = b""
        r.headers = headers or {}
        return r

    # Retry-After header missing -> defaults to default_delay_sec (1.0)
    mock_post.reset_mock()
    mock_sleep.reset_mock()
    mock_post.side_effect = [
        _resp(requests.codes.too_many_requests, {}),
        _resp(requests.codes.no_content, {}),
    ]

    assert obj.send(body="test") is True
    assert mock_post.call_count == 2
    assert mock_sleep.call_count == 1
    assert mock_sleep.call_args_list[-1][0][0] == pytest.approx(
        1.0, rel=0, abs=0.05
    )

    # Retry-After invalid -> falls back to 1.0
    mock_post.reset_mock()
    mock_sleep.reset_mock()
    mock_post.side_effect = [
        _resp(requests.codes.too_many_requests, {"Retry-After": "garbage"}),
        _resp(requests.codes.no_content, {}),
    ]

    assert obj.send(body="test") is True
    assert mock_post.call_count == 2
    assert mock_sleep.call_count >= 1
    assert mock_sleep.call_args_list[-1][0][0] == pytest.approx(
        1.0, rel=0, abs=0.05
    )

    # Retry-After < 1.0 -> max(1.0, value) enforces 1.0
    mock_post.reset_mock()
    mock_sleep.reset_mock()
    mock_post.side_effect = [
        _resp(requests.codes.too_many_requests, {"Retry-After": "0.25"}),
        _resp(requests.codes.no_content, {}),
    ]

    assert obj.send(body="test") is True
    assert mock_post.call_count == 2
    assert mock_sleep.call_count >= 1
    assert mock_sleep.call_args_list[-1][0][0] == pytest.approx(
        1.0, rel=0, abs=0.05
    )

    # Retry-After valid integer -> sleeps that many seconds
    mock_post.reset_mock()
    mock_sleep.reset_mock()
    mock_post.side_effect = [
        _resp(requests.codes.too_many_requests, {"Retry-After": "2"}),
        _resp(requests.codes.no_content, {}),
    ]

    assert obj.send(body="test") is True
    assert mock_post.call_count == 2
    assert mock_sleep.call_count >= 1
    assert mock_sleep.call_args_list[-1][0][0] == pytest.approx(
        2.0, rel=0, abs=0.05
    )

    # Retry exhaustion: default send() retries once.
    # If we get 429 twice, second one is not retried and send fails.
    mock_post.reset_mock()
    mock_sleep.reset_mock()
    mock_post.side_effect = [
        _resp(requests.codes.too_many_requests, {"Retry-After": "1"}),
        _resp(requests.codes.too_many_requests, {"Retry-After": "1"}),
    ]

    assert obj.send(body="test") is False
    assert mock_post.call_count == 2
    assert mock_sleep.call_count >= 1

    mock_sleep.reset_mock()
    mock_post.reset_mock()

    response = mock.Mock()
    response.status_code = requests.codes.no_content
    response.content = b""
    response.headers = {}
    mock_post.return_value = response
    mock_post.side_effect = None

    # Force the 'now <= ratelimit_reset' path to compute a wait.
    obj = NotifyFluxer(webhook_id=webhook_id, webhook_token=webhook_token)

    # Force the rate-limit gate to run
    obj.ratelimit_remaining = 0.0

    now = datetime.now(timezone.utc).replace(tzinfo=None)
    obj.ratelimit_reset = now - timedelta(seconds=2)

    with mock.patch.object(obj, "throttle") as m_throttle:
        assert obj.send(body="test") is True

    # We expect throttle(wait=~2.0) to be called.
    assert m_throttle.call_count >= 1
    wait = m_throttle.call_args_list[-1][1].get("wait")

    assert wait is None

    # Call count
    assert mock_post.call_count == 1

    # Reset our object
    mock_post.reset_mock()

    # Force the 'now < ratelimit_reset' path to compute a wait.
    obj = NotifyFluxer(webhook_id=webhook_id, webhook_token=webhook_token)

    # Force the rate-limit gate to run
    obj.ratelimit_remaining = 0.0

    # Bracket the call with timestamps so the assertion is
    # architecture-independent (slow ppc64le build machines can add
    # hundreds of ms of overhead between captures of datetime.now()).
    t_before = datetime.now(timezone.utc).replace(tzinfo=None)
    obj.ratelimit_reset = t_before + timedelta(seconds=2)

    with mock.patch.object(obj, "throttle") as m_throttle:
        assert obj.send(body="test") is True

    t_after = datetime.now(timezone.utc).replace(tzinfo=None)

    # We expect throttle(wait=~2.0) to be called.
    assert m_throttle.call_count >= 1
    wait = m_throttle.call_args_list[-1][1].get("wait")

    assert wait is not None
    # wait = ratelimit_reset - now_inside_plugin
    # The plugin's 'now' is somewhere in [t_before, t_after], so:
    #   lower bound: ratelimit_reset - t_after = 2.0 - elapsed
    #   upper bound: ratelimit_reset - t_before = 2.0
    elapsed = (t_after - t_before).total_seconds()
    assert (2.0 - elapsed) <= wait <= 2.0

    # Call count
    assert mock_post.call_count == 1


@mock.patch("requests.post")
def test_plugin_fluxer_general(
    mock_post: mock.MagicMock,
) -> None:
    """NotifyFluxer() General Checks."""

    webhook_id, webhook_token = _tokens()

    # Prepare Mock
    mock_post.return_value = requests.Request()
    mock_post.return_value.status_code = requests.codes.ok
    mock_post.return_value.content = ""

    # Invalid webhook id
    with pytest.raises(TypeError):
        NotifyFluxer(webhook_id=None, webhook_token=webhook_token)
    # Invalid webhook id (whitespace)
    with pytest.raises(TypeError):
        NotifyFluxer(webhook_id="  ", webhook_token=webhook_token)

    # Invalid webhook token
    with pytest.raises(TypeError):
        NotifyFluxer(webhook_id=webhook_id, webhook_token=None)

    # Private mode but no hostname provided
    with pytest.raises(TypeError):
        NotifyFluxer(
            webhook_id=webhook_id,
            webhook_token=webhook_token,
            mode="private",
        )

    obj = NotifyFluxer(
        webhook_id=webhook_id,
        webhook_token=webhook_token,
        footer=True,
        include_image=True,
    )

    # Simple Markdown Single line of text
    test_markdown = "body"
    desc, results = obj.extract_markdown_sections(test_markdown)
    assert isinstance(results, list) is True
    assert len(results) == 0
    assert desc == "body"

    # Test our header parsing when not lead with a header
    test_markdown = """
    A section of text that has no header at the top.
    It also has a hash tag # <- in the middle of a
    string.

    ## Heading 1
    body

    # Heading 2

    more content
    on multi-lines
    """

    desc, results = obj.extract_markdown_sections(test_markdown)
    # we have a description
    assert isinstance(desc, str) is True
    assert desc.startswith("A section of text that has no header at the top.")
    assert desc.endswith("string.")

    assert isinstance(results, list) is True
    assert len(results) == 2
    assert results[0]["name"] == "Heading 1"
    assert results[0]["value"] == "```md\nbody\n```"
    assert results[1]["name"] == "Heading 2"
    assert (
        results[1]["value"] == "```md\nmore content\n    on multi-lines\n```"
    )

    # Test our header parsing
    test_markdown = (
        "## Heading one\nbody body\n\n"
        + "# Heading 2 ##\n\nTest\n\n"
        + "more content\n"
        + "even more content  \t\r\n\n\n"
        + "# Heading 3 ##\n\n\n"
        + "normal content\n"
        + "# heading 4\n"
        + "#### Heading 5"
    )

    desc, results = obj.extract_markdown_sections(test_markdown)
    assert isinstance(results, list) is True
    # No desc details filled out
    assert isinstance(desc, str) is True
    assert not desc

    # We should have 5 sections (since there are 5 headers identified above)
    assert len(results) == 5
    assert results[0]["name"] == "Heading one"
    assert results[0]["value"] == "```md\nbody body\n```"
    assert results[1]["name"] == "Heading 2"
    assert (
        results[1]["value"]
        == "```md\nTest\n\nmore content\neven more content\n```"
    )
    assert results[2]["name"] == "Heading 3"
    assert results[2]["value"] == "```md\nnormal content\n```"
    assert results[3]["name"] == "heading 4"
    assert results[3]["value"] == "```\n```"
    assert results[4]["name"] == "Heading 5"
    assert results[4]["value"] == "```\n```"

    # Create an apprise instance
    a = Apprise()

    # Our processing is slightly different when we aren't using markdown
    # as we do not pre-parse content during our notifications
    assert (
        a.add(
            f"fluxer://{webhook_id}/{webhook_token}/"
            "?format=markdown&footer=Yes"
        )
        is True
    )

    # This call includes an image with its payload:
    orig_fluxer_max_fields = NotifyFluxer.fluxer_max_fields
    try:
        NotifyFluxer.fluxer_max_fields = 1
        assert (
            a.notify(
                body=test_markdown,
                title="title",
                notify_type=NotifyType.INFO,
                body_format=NotifyFormat.TEXT,
            )
            is True
        )
    finally:
        # Restore the original value to avoid impacting other tests
        NotifyFluxer.fluxer_max_fields = orig_fluxer_max_fields

    # Throw an exception on the forth call to requests.post()
    # This allows to test our batch field processing
    response = mock.Mock()
    response.content = b""
    response.status_code = requests.codes.ok
    response.headers = {}
    mock_post.return_value = response
    mock_post.side_effect = [
        response,
        response,
        response,
        requests.RequestException(),
    ]

    # Test our markdown
    obj = Apprise.instantiate(
        f"fluxer://{webhook_id}/{webhook_token}/?format=markdown"
    )
    assert isinstance(obj, NotifyFluxer)

    # Force batching so we actually hit the 4th post() call
    orig_fluxer_max_fields = NotifyFluxer.fluxer_max_fields
    try:
        NotifyFluxer.fluxer_max_fields = 1

        assert (
            obj.notify(
                body=test_markdown, title="title", notify_type=NotifyType.INFO
            )
            is False
        )
    finally:
        NotifyFluxer.fluxer_max_fields = orig_fluxer_max_fields
        mock_post.side_effect = None

    # Empty String
    desc, results = obj.extract_markdown_sections("")
    assert isinstance(results, list) is True
    assert len(results) == 0

    # No desc details filled out
    assert isinstance(desc, str) is True
    assert not desc

    # String without Heading
    test_markdown = (
        "Just a string without any header entries.\n" + "A second line"
    )
    desc, results = obj.extract_markdown_sections(test_markdown)
    assert isinstance(results, list) is True
    assert len(results) == 0

    # No desc details filled out
    assert isinstance(desc, str) is True
    assert (
        desc == "Just a string without any header entries.\n" + "A second line"
    )

    # Use our test markdown string during a notification
    assert (
        obj.notify(
            body=test_markdown, title="title", notify_type=NotifyType.INFO
        )
        is True
    )

    # Create an apprise instance
    a = Apprise()

    # Our processing is slightly different when we aren't using markdown
    # as we do not pre-parse content during our notifications
    assert (
        a.add(
            f"fluxer://{webhook_id}/{webhook_token}/"
            "?format=markdown&footer=Yes"
        )
        is True
    )

    # This call includes an image with it's payload:
    assert (
        a.notify(
            body=test_markdown,
            title="title",
            notify_type=NotifyType.INFO,
            body_format=NotifyFormat.TEXT,
        )
        is True
    )

    assert (
        a.notify(
            body=test_markdown,
            title="title",
            notify_type=NotifyType.INFO,
            body_format=NotifyFormat.MARKDOWN,
        )
        is True
    )

    # Toggle our logo availability
    a.asset.image_url_logo = None
    assert (
        a.notify(body="body", title="title", notify_type=NotifyType.INFO)
        is True
    )

    # Create an apprise instance
    a = Apprise()

    # Reset our object
    mock_post.reset_mock()

    # Test our threading
    assert (
        a.add(f"fluxer://{webhook_id}/{webhook_token}/?thread=12345") is True
    )

    # This call includes an image with it's payload:
    assert a.notify(body="test", title="title") is True

    assert mock_post.call_count == 1
    response = mock_post.call_args_list[0][1]
    assert "params" in response
    assert response["params"].get("thread_id") == "12345"


@mock.patch("requests.post")
def test_plugin_fluxer_overflow(mock_post):
    """NotifyFluxer() Overflow Checks."""

    webhook_id, webhook_token = _tokens()

    # Prepare Mock
    mock_post.return_value = requests.Request()
    mock_post.return_value.status_code = requests.codes.ok

    # Some variables we use to control the data we work with
    body_len = 8000
    title_len = 1024

    # Number of characters per line
    row = 24

    # Create a large body and title with random data
    body = "".join(choice(str_alpha + str_num + " ") for _ in range(body_len))
    body = "\r\n".join([body[i : i + row] for i in range(0, len(body), row)])

    # Create our title using random data
    title = "".join(choice(str_alpha + str_num) for _ in range(title_len))

    results = NotifyFluxer.parse_url(
        f"fluxer://{webhook_id}/{webhook_token}/?overflow=split"
    )

    instance = NotifyFluxer(**results)
    assert isinstance(instance, NotifyFluxer)

    results = instance._apply_overflow(
        body, title=title, overflow=OverflowMode.SPLIT
    )

    # Ensure we never exceed 2000 characters
    for entry in results:
        assert len(entry["title"]) <= instance.title_maxlen
        assert len(entry["title"]) + len(entry["body"]) <= instance.body_maxlen


@mock.patch("requests.post")
def test_plugin_fluxer_markdown_extra(mock_post):
    """NotifyFluxer() Markdown Extra Checks."""

    webhook_id, webhook_token = _tokens()

    # Prepare Mock
    mock_post.return_value = requests.Request()
    mock_post.return_value.status_code = requests.codes.ok
    mock_post.return_value.content = b""
    mock_post.return_value.headers = {}

    # Reset our apprise object
    a = Apprise()

    assert (
        a.add(
            f"fluxer://{webhook_id}/{webhook_token}/"
            "?format=markdown&footer=Yes"
        )
        is True
    )

    test_markdown = "[green-blue](https://google.com)"

    # This call includes an image with it's payload:
    assert (
        a.notify(
            body=test_markdown,
            title="title",
            notify_type=NotifyType.INFO,
            body_format=NotifyFormat.TEXT,
        )
        is True
    )

    assert (
        a.notify(body="body", title="title", notify_type=NotifyType.INFO)
        is True
    )


@mock.patch("requests.post")
def test_plugin_fluxer_markdown_attachments(
    mock_post: mock.MagicMock,
) -> None:
    # Prepare our tokens
    webhook_id, webhook_token = _tokens()

    # Prepare a good response
    response = mock.Mock()
    response.status_code = requests.codes.ok
    response.content = b""

    # Prepare a bad response
    bad_response = mock.Mock()
    bad_response.status_code = requests.codes.internal_server_error
    bad_response.content = b""

    # Prepare Mock return object
    mock_post.return_value = response

    # Test our markdown
    obj = Apprise.instantiate(
        f"fluxer://{webhook_id}/{webhook_token}/?format=markdown"
    )

    # attach our content
    attach = AppriseAttachment(os.path.join(TEST_VAR_DIR, "apprise-test.gif"))

    assert (
        obj.notify(
            body="body",
            title="title",
            notify_type=NotifyType.INFO,
            attach=attach,
        )
        is True
    )

    # Test our call count
    assert mock_post.call_count == 2
    assert (
        mock_post.call_args_list[0][0][0]
        == f"https://api.fluxer.app/webhooks/{webhook_id}/{webhook_token}"
    )
    assert (
        mock_post.call_args_list[1][0][0]
        == f"https://api.fluxer.app/webhooks/{webhook_id}/{webhook_token}"
    )

    # Reset our object
    mock_post.reset_mock()

    # Test notifications with mentions and attachments in it
    assert (
        obj.notify(
            body="Say hello to <@1234>!",
            notify_type=NotifyType.INFO,
            attach=attach,
        )
        is True
    )

    # Test our call count
    assert mock_post.call_count == 2
    assert (
        mock_post.call_args_list[0][0][0]
        == f"https://api.fluxer.app/webhooks/{webhook_id}/{webhook_token}"
    )
    assert (
        mock_post.call_args_list[1][0][0]
        == f"https://api.fluxer.app/webhooks/{webhook_id}/{webhook_token}"
    )

    # Reset our object
    mock_post.reset_mock()

    # An invalid attachment will cause a failure
    path = os.path.join(TEST_VAR_DIR, "/invalid/path/to/an/invalid/file.jpg")
    attach = AppriseAttachment(path)
    assert (
        obj.notify(
            body="body",
            title="title",
            notify_type=NotifyType.INFO,
            attach=path,
        )
        is False
    )

    # update our attachment to be valid
    attach = AppriseAttachment(os.path.join(TEST_VAR_DIR, "apprise-test.gif"))

    mock_post.return_value = None
    # Throw an exception on the first call to requests.post()
    for side_effect in (requests.RequestException(), OSError(), bad_response):
        mock_post.side_effect = [side_effect]

        # We'll fail now because of our error handling
        assert obj.send(body="test", attach=attach) is False

    # Throw an exception on the second call to requests.post()
    for side_effect in (requests.RequestException(), OSError(), bad_response):
        mock_post.side_effect = [response, side_effect]

        # We'll fail now because of our error handling
        assert obj.send(body="test", attach=attach) is False

    # handle a bad response
    bad_response = mock.Mock()
    bad_response.content = b""
    bad_response.headers = {}
    bad_response.status_code = requests.codes.internal_server_error
    mock_post.side_effect = [response, bad_response]

    # We'll fail now because of an internal exception
    assert obj.send(body="test", attach=attach) is False


@mock.patch("requests.post")
def test_plugin_fluxer_markdown_fields_batches_exactly(mock_post):
    webhook_id, webhook_token = _tokens()
    response = mock.Mock()
    response.status_code = requests.codes.ok
    response.content = ""
    response.headers = {}
    mock_post.return_value = response

    # Force tiny batches
    NotifyFluxer.fluxer_max_fields = 1

    body = "# H1\nv1\n# H2\nv2\n# H3\nv3\n"
    obj = Apprise.instantiate(
        f"fluxer://{webhook_id}/{webhook_token}/?format=markdown&fields=yes"
    )
    assert isinstance(obj, NotifyFluxer)

    assert obj.send(body=body) is True

    # H1, H2, H3 => 3 fields => 3 posts (since max_fields=1)
    assert mock_post.call_count == 3

    NotifyFluxer.fluxer_max_fields = 10


@mock.patch("requests.post")
def test_plugin_fluxer_markdown_ping_is_additive(
    mock_post: mock.MagicMock,
) -> None:
    webhook_id, webhook_token = _tokens()

    mock_post.return_value = requests.Request()
    mock_post.return_value.status_code = requests.codes.ok
    mock_post.return_value.content = b""
    mock_post.return_value.headers = {}

    body = "Body pings <@111> and <@&222> @everyone"
    results = NotifyFluxer.parse_url(
        f"fluxer://{webhook_id}/{webhook_token}/"
        "?format=markdown"
        "&ping=<@333>,<@%26444>,@joe"
    )
    obj = NotifyFluxer(**results)

    assert obj.send(body=body) is True
    assert mock_post.call_count == 1

    payload = loads(mock_post.call_args_list[0][1]["data"])

    assert "allow_mentions" in payload
    # union
    assert set(payload["allow_mentions"]["users"]) == {"111", "333"}
    assert set(payload["allow_mentions"]["roles"]) == {"222", "444"}
    assert set(payload["allow_mentions"]["parse"]) == {"everyone", "joe"}
    assert payload["content"].startswith("👉 ")


@mock.patch("requests.post")
def test_plugin_fluxer_html_ping_is_exclusive(
    mock_post: mock.MagicMock,
) -> None:
    webhook_id, webhook_token = _tokens()

    mock_post.return_value = requests.Request()
    mock_post.return_value.status_code = requests.codes.ok

    body = "Body includes <@111> <@&222> @everyone but must be ignored"
    results = NotifyFluxer.parse_url(
        f"fluxer://{webhook_id}/{webhook_token}/"
        "?format=html"
        "&ping=<@333>,<@%26444>,@joe"
    )
    obj = NotifyFluxer(**results)

    assert obj.send(body=body) is True
    payload = loads(mock_post.call_args_list[0][1]["data"])

    assert set(payload["allow_mentions"]["users"]) == {"333"}
    assert set(payload["allow_mentions"]["roles"]) == {"444"}
    assert set(payload["allow_mentions"]["parse"]) == {"joe"}


@mock.patch("requests.post")
def test_plugin_fluxer_markdown_no_mentions_has_no_allow_mentions(
    mock_post: mock.MagicMock,
) -> None:
    webhook_id, webhook_token = _tokens()

    mock_post.return_value = requests.Request()
    mock_post.return_value.status_code = requests.codes.ok
    mock_post.return_value.content = b""
    mock_post.return_value.headers = {}

    results = NotifyFluxer.parse_url(
        f"fluxer://{webhook_id}/{webhook_token}/?format=markdown"
    )
    obj = NotifyFluxer(**results)

    assert obj.send(body="Hello world") is True
    payload = loads(mock_post.call_args_list[0][1]["data"])

    assert "allow_mentions" not in payload
    assert "content" not in payload


@mock.patch("requests.post")
def test_plugin_fluxer_markdown_single_field_posts_once(
    mock_post: mock.MagicMock,
) -> None:
    webhook_id, webhook_token = _tokens()

    response = mock.Mock()
    response.status_code = requests.codes.ok
    response.content = b""
    response.headers = {}
    mock_post.return_value = response

    NotifyFluxer.fluxer_max_fields = 10

    body = "# H1\nv1\n"
    obj = Apprise.instantiate(
        f"fluxer://{webhook_id}/{webhook_token}/?format=markdown&fields=yes"
    )
    assert isinstance(obj, NotifyFluxer)

    assert obj.send(body=body) is True
    assert mock_post.call_count == 1


@mock.patch("requests.post")
def test_plugin_fluxer_threading(mock_post: mock.MagicMock) -> None:
    """Threading passes thread_id parameter."""

    webhook_id, webhook_token = _tokens()

    response = mock.Mock()
    response.status_code = requests.codes.ok
    response.content = b""
    response.headers = {}
    mock_post.return_value = response

    a = Apprise()
    assert (
        a.add(
            f"fluxer://{webhook_id}/{webhook_token}/"
            "?thread=12345&thread_name=abc"
        )
        is True
    )

    assert a.notify(body="test", title="title") is True

    kwargs = mock_post.call_args_list[0][1]
    assert "params" in kwargs
    assert kwargs["params"].get("thread_id") == "12345"

    payload = loads(mock_post.call_args_list[0][1]["data"])
    assert payload.get("thread_name") == "abc"

    assert "thread_name=abc" in a[0].url()
    assert "thread=12345" in a[0].url()


@mock.patch("requests.post")
@mock.patch("time.sleep")
@mock.patch("builtins.open")
def test_plugin_fluxer_429_attachment_closes_edge_cases(
    mock_open: mock.MagicMock,
    mock_sleep: mock.MagicMock,
    mock_post: mock.MagicMock,
) -> None:
    """
    Cover 429 error during attachment upload triggers the pre-recursion close
    loop, and close() exceptions are suppressed.
    """

    mock_sleep.return_value = True

    webhook_id, webhook_token = _tokens()
    obj = NotifyFluxer(webhook_id=webhook_id, webhook_token=webhook_token)

    attach_path = os.path.join(TEST_VAR_DIR, "apprise-test.png")
    attach = AppriseAttachment(attach_path)

    class _BadCloseIO:
        def __init__(self) -> None:
            self.closed = False

        def close(self) -> None:
            self.closed = True
            raise OSError("boom")

    bad = _BadCloseIO()
    mock_open.return_value = bad

    def _resp(code: int, headers: dict[str, str] | None = None) -> mock.Mock:
        r = mock.Mock()
        r.status_code = code
        r.content = b""
        r.headers = headers or {}
        return r

    # 3 posts:
    #  (1) initial message post succeeds
    #  (2) attachment post 429 triggers file close-before-recursion
    #  (3) recursive retry succeeds (note: retry is non-attachment in baseline)
    mock_post.side_effect = [
        _resp(requests.codes.no_content, {}),
        _resp(requests.codes.too_many_requests, {"Retry-After": "1"}),
        _resp(requests.codes.no_content, {}),
    ]

    assert obj.send(body="test", attach=attach) is True
    assert bad.closed is True

    mock_post.reset_mock()
    mock_open.reset_mock()

    mock_sleep.reset_mock()
    mock_sleep.return_value = True

    webhook_id, webhook_token = _tokens()
    obj = NotifyFluxer(webhook_id=webhook_id, webhook_token=webhook_token)

    attach_path = os.path.join(TEST_VAR_DIR, "apprise-test.png")
    attach = AppriseAttachment(attach_path)

    class _GoodCloseIO:
        def __init__(self) -> None:
            self.closed = False

        def close(self) -> None:
            self.closed = True

    good = _GoodCloseIO()
    mock_open.return_value = good

    def _resp(code: int, headers: dict[str, str] | None = None) -> mock.Mock:
        r = mock.Mock()
        r.status_code = code
        r.content = b""
        r.headers = headers or {}
        return r

    mock_post.side_effect = [
        _resp(requests.codes.no_content, {}),
        _resp(requests.codes.too_many_requests, {"Retry-After": "1"}),
        _resp(requests.codes.no_content, {}),
    ]

    assert obj.send(body="test", attach=attach) is True
    assert good.closed is True


@mock.patch("requests.post")
def test_plugin_fluxer_attach_memory(mock_post: mock.MagicMock) -> None:
    """Regression: AttachMemory must be sendable without OSError."""
    from apprise.attachment.memory import AttachMemory

    webhook_id, webhook_token = _tokens()

    response = mock.Mock()
    response.status_code = requests.codes.ok
    response.content = b""
    mock_post.return_value = response

    obj = Apprise.instantiate(f"fluxer://{webhook_id}/{webhook_token}/")

    mem = AttachMemory(
        content=b"<html><body><h1>Test</h1></body></html>",
        name="report.html",
        mimetype="text/html",
    )

    assert obj.notify(body="Test", attach=mem) is True
    assert mock_post.call_count >= 1
